"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.connection.api_gateway_connection import \
    REQ_VALIDATOR_PARAM_NAME, \
    REQ_VALIDATOR_PARAM_VALIDATE_BODY, REQ_VALIDATOR_PARAM_VALIDATE_PARAMS, \
    ApiGatewayConnection
from syndicate.core.resources.api_gateway_resource import API_REQUIRED_PARAMS
from syndicate.core.resources.helper import validate_params
from syndicate.core.transform.terraform.converter.tf_resource_converter import \
    TerraformResourceConverter
from syndicate.core.transform.terraform.tf_resource_name_builder import \
    build_terraform_resource_name
from syndicate.core.transform.terraform.tf_resource_reference_builder import \
    build_api_gateway_resource_id_ref, build_api_gateway_resource_path_ref
from syndicate.core.transform.terraform.tf_resource_reference_builder import \
    build_function_invoke_arn_ref, build_authorizer_id_ref, \
    build_api_gateway_method_name_reference, build_function_name_ref, \
    build_rest_api_id_ref, build_api_gateway_root_resource_id_ref, \
    build_api_gateway_deployment_id_ref

NONE_AUTH = 'NONE'
CUSTOM_AUTH = 'CUSTOM'
AWS_IAM_AUTH = 'AWS_IAM'

API_GATEWAY_SUPPORTED_METHODS = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE',
                                 'OPTIONS',
                                 'HEAD', 'ANY']


class ApiGatewayConverter(TerraformResourceConverter):

    def convert(self, name, resource):
        validate_params(name, resource, API_REQUIRED_PARAMS)

        api_name = name
        rest_api_template = generate_tf_template_for_api_gateway(
            api_name=api_name)
        self.template.add_aws_api_gateway_rest_api(meta=rest_api_template)
        auth_mappings = self._transform_authorizers(resource=resource,
                                                    api_name=api_name)

        method_names = []
        integration_names = []
        resources = resource.get('resources')
        for res_name, res in resources.items():
            api_gateway_resource, resource_name = self._create_api_gateway_resource(
                path_part=res_name, api_name=api_name)

            for http_method in API_GATEWAY_SUPPORTED_METHODS:
                method_meta = res.get(http_method)
                if method_meta:
                    tf_method_resource_name = build_terraform_resource_name(
                        resource_name, http_method)
                    method_names.append(tf_method_resource_name)
                    self.create_request_validator(method_meta=method_meta,
                                                  api_name=api_name,
                                                  resource_name=resource_name,
                                                  http_method=http_method)

                    authorizer_id = None
                    authorization_type = method_meta.get('authorization_type')
                    if authorization_type not in ['NONE', 'AWS_IAM']:
                        authorizer_id = auth_mappings.get(
                            authorization_type)
                        if not authorizer_id:
                            raise AssertionError(
                                'Authorizer {0} does not exist'.format(
                                    authorization_type))
                        authorization_type = 'CUSTOM'

                    method_request_parameters = method_meta.get(
                        'method_request_parameters')
                    method_template = get_api_gateway_method(
                        http_method=http_method,
                        resource_name=resource_name,
                        rest_api=api_name,
                        authorization=authorization_type,
                        method_name=tf_method_resource_name,
                        request_parameters=method_request_parameters,
                        authorizer_id=authorizer_id)
                    self.template.add_aws_api_gateway_method(
                        meta=method_template)

                    integration_type = method_meta.get('integration_type')
                    integration_name = build_terraform_resource_name(
                        resource_name, http_method)
                    integration_names.append(integration_name)
                    if integration_type:
                        if integration_type == 'lambda':
                            self._create_lambda_integration(
                                method_meta=method_meta,
                                method_name=tf_method_resource_name,
                                api_name=api_name,
                                integration_name=integration_name,
                                resource_name=resource_name)
                        elif integration_type == 'service':
                            self._create_service_integration(
                                resource_name=resource_name,
                                method_meta=method_meta,
                                method_name=tf_method_resource_name,
                                api_name=api_name,
                                integration_name=integration_name)
                        elif integration_type == 'mock':
                            self._create_mock_integration(
                                method_meta=method_meta,
                                resource_name=resource_name, api_name=api_name,
                                integration_name=integration_name,
                                method_name=tf_method_resource_name)
                        elif integration_type == 'http':
                            self._create_http_integration(
                                method_meta=method_meta,
                                method_name=tf_method_resource_name,
                                integration_name=integration_name,
                                resource_name=resource_name,
                                api_name=api_name)

                    responses = method_meta.get('responses')
                    for response in responses:
                        status_code = response.get('status_code')
                        method_response = api_gateway_method_response(
                            resource_name=resource_name,
                            status_code=status_code,
                            api_name=api_name,
                            http_method=http_method,
                            method_name=tf_method_resource_name)
                        self.template.add_aws_api_gateway_method_response(
                            meta=method_response)

                    integration_responses = method_meta.get(
                        'integration_responses')
                    for int_response in integration_responses:
                        status_code = int_response.get('status_code')
                        response_templates = int_response.get(
                            'response_templates')
                        error_regex = int_response.get('error_regex')
                        integration_response = create_api_gateway_integration_response(
                            resource_name=resource_name,
                            api_name=api_name,
                            status_code=status_code,
                            response_template=response_templates,
                            http_method=http_method,
                            method_name=tf_method_resource_name,
                            integration=integration_name,
                            selection_pattern=error_regex)
                        self.template.add_aws_api_gateway_integration_response(
                            meta=integration_response)

        self._create_api_gateway_deployment(resource=resource,
                                            api_name=api_name,
                                            method_names=method_names,
                                            integration_names=integration_names)

    def _create_api_gateway_deployment(self, resource, api_name, method_names,
                                       integration_names):
        deploy_stage = resource.get('deploy_stage')
        if deploy_stage:
            deployment_name = build_terraform_resource_name(api_name,
                                                            deploy_stage,
                                                            'deployment')
            deployment = api_gateway_deployment(api_name=api_name,
                                                deployment_name=deployment_name,
                                                methods=method_names,
                                                integration_names=integration_names)
            cache_cluster_configuration = resource.get(
                'cluster_cache_configuration')
            stage = api_gateway_stage(api_name=api_name,
                                      stage_name=deploy_stage,
                                      deployment_name=deployment_name,
                                      cache_cluster_configuration=cache_cluster_configuration)
            self.template.add_aws_api_gateway_stage(meta=stage)
            self.template.add_aws_api_gateway_deployment(meta=deployment)

    def _create_api_gateway_resource(self, path_part, api_name):
        resource = None
        resource_name = None
        parent_id = build_api_gateway_root_resource_id_ref(api_name=api_name)
        resource_path = path_part[1:]
        path_arr = resource_path.split('/')
        for path in path_arr:
            cur_resource_path = ''.join(path_arr[:path_arr.index(path) + 1])
            api_resource_name = build_terraform_resource_name(
                cur_resource_path, 'resource').replace('_', '-')
            existing_resource = self.template.get_resource_by_name(
                api_resource_name)
            if existing_resource:
                resource = existing_resource
                resource_name = api_resource_name
                continue
            api_gateway_resource = get_api_gateway_resource(
                path_part=path,
                rest_api=api_name,
                resource_name=api_resource_name,
                parent_id=parent_id)
            self.template.add_aws_api_gateway_resource(
                meta=api_gateway_resource)
            parent_id = build_api_gateway_resource_id_ref(
                resource_name=api_resource_name)
            resource = api_gateway_resource
            resource_name = api_resource_name
        return resource, resource_name

    def _transform_authorizers(self, resource, api_name):
        authorizer_mappings = {}
        authorizers = resource.get('authorizers', {})
        for name, val in authorizers.items():
            lambda_arn = self.define_function_arn(meta=val)

            uri = 'arn:aws:apigateway:{0}:lambda:path/2015-03-31/' \
                  'functions/{1}/invocations'.format(self.config.region,
                                                     lambda_arn)
            identity_source = val.get('identity_source')
            ttl = val.get('ttl')
            auth_type = val.get('type')

            authorizer_name = f'{api_name}_authorizer'
            authorizer_ref = build_authorizer_id_ref(
                authorizer_name=authorizer_name)
            authorizer = api_gateway_authorizer(
                authorizer_name=authorizer_name,
                authorizer_uri=uri,
                rest_api_name=api_name,
                ttl=ttl,
                identity_source=identity_source,
                auth_type=auth_type)
            self.template.add_aws_api_gateway_authorizer(meta=authorizer)
            authorizer_mappings.update({name: authorizer_ref})
        return authorizer_mappings

    def define_function_arn(self, meta):
        lambda_name = meta.get('lambda_name')

        lambda_meta = self.template.get_resource_by_name(lambda_name)
        if lambda_meta:
            return build_function_invoke_arn_ref(lambda_name)
        lambda_service = self.resources_provider.lambda_resource().lambda_conn

        function = lambda_service.get_function(lambda_name=lambda_name)
        if not function:
            raise AssertionError(
                f'Specified lambda {lambda_name} does not exists')

        version = meta.get('lambda_version')
        alias = meta.get('lambda_alias')
        if version or alias:
            return self.build_lambda_arn_with_alias(function, alias)
        else:
            return function['Configuration']['FunctionArn']

    def build_lambda_arn_with_alias(self, response, alias=None):
        name = response['Configuration']['FunctionName']
        l_arn = self.build_lambda_arn(name=name)
        version = response['Configuration']['Version']
        arn = '{0}:{1}'.format(l_arn, version)
        # override version if alias exists
        if alias:
            arn = '{0}:{1}'.format(l_arn, alias)
        return arn

    def build_lambda_arn(self, name):
        region = self.config.region
        account_id = self.config.account_id
        return f'arn:aws:lambda:{region}:{account_id}:function:{name}'

    def _create_lambda_integration(self, method_meta, method_name, api_name,
                                   integration_name, resource_name):
        enable_proxy = method_meta.get('enable_proxy')
        passthrough_behavior = method_meta.get(
            'integration_passthrough_behavior')
        integration_request_template = method_meta.get(
            'integration_request_body_template')

        cache_configuration = method_meta.get(
            'cache_configuration')
        cache_key_parameters = cache_configuration.get(
            'cache_key_parameters') if cache_configuration else None

        int_type = 'AWS_PROXY' if enable_proxy else 'AWS'
        uri = self.define_function_arn(meta=method_meta)

        passthrough_behavior = passthrough_behavior if passthrough_behavior else 'WHEN_NO_MATCH'
        integration = create_api_gateway_integration(
            integration_name=integration_name,
            api_name=api_name,
            method_name=method_name,
            resource_name=resource_name,
            integration_type=int_type,
            request_template=integration_request_template,
            integration_method='POST',
            uri=uri,
            passthrough_behavior=passthrough_behavior,
            cache_key_parameters=cache_key_parameters)
        self.template.add_aws_api_gateway_integration(
            meta=integration)

    def _create_service_integration(self, method_meta, integration_name,
                                    api_name, method_name, resource_name):
        integration_request_template = method_meta.get(
            'integration_request_body_template')
        passthrough_behavior = method_meta.get(
            'integration_passthrough_behavior')
        uri = method_meta.get('uri')
        role = method_meta.get('role')
        integration_method = method_meta.get(
            'integration_method')
        uri = 'arn:aws:apigateway:{0}'.format(uri)

        credentials = ApiGatewayConnection.get_service_integration_credentials(
            self.config.accountId, role)
        passthrough_behavior = passthrough_behavior if passthrough_behavior else 'WHEN_NO_MATCH'

        integration = create_api_gateway_integration(
            integration_name=integration_name,
            api_name=api_name,
            method_name=method_name,
            resource_name=resource_name,
            integration_type='AWS',
            request_template=integration_request_template,
            integration_method=integration_method,
            uri=uri,
            passthrough_behavior=passthrough_behavior,
            credentials=credentials)
        self.template.add_aws_api_gateway_integration(
            meta=integration)

    def _create_mock_integration(self, method_meta, integration_name, api_name,
                                 method_name, resource_name):
        integration_request_template = method_meta.get(
            'integration_request_body_template')
        passthrough_behavior = method_meta.get(
            'integration_passthrough_behavior')

        passthrough_behavior = passthrough_behavior if passthrough_behavior else 'WHEN_NO_MATCH'
        integration = create_api_gateway_integration(
            integration_name=integration_name,
            api_name=api_name,
            method_name=method_name,
            resource_name=resource_name,
            integration_type='MOCK',
            request_template=integration_request_template,
            passthrough_behavior=passthrough_behavior)
        self.template.add_aws_api_gateway_integration(
            meta=integration)

    def _create_http_integration(self, method_meta, integration_name, api_name,
                                 method_name, resource_name):
        integration_request_template = method_meta.get(
            'integration_request_body_template')
        passthrough_behavior = method_meta.get(
            'integration_passthrough_behavior')

        integration_method = method_meta.get(
            'integration_method')
        uri = method_meta.get('uri')
        enable_proxy = method_meta.get('enable_proxy')

        int_type = 'HTTP_PROXY' if enable_proxy else 'HTTP'
        passthrough_behavior = passthrough_behavior if passthrough_behavior else 'WHEN_NO_MATCH'
        integration = create_api_gateway_integration(
            integration_name=integration_name,
            api_name=api_name,
            method_name=method_name,
            resource_name=resource_name,
            integration_type=int_type,
            request_template=integration_request_template,
            integration_method=integration_method,
            uri=uri,
            passthrough_behavior=passthrough_behavior)
        self.template.add_aws_api_gateway_integration(
            meta=integration)

    def _get_api_source_arn(self, rest_api_name, method_name, resource_name):
        rest_api_id = build_rest_api_id_ref(api_name=rest_api_name)
        region = '${' + self.config.region + '}'
        account_id = '${' + self.config.accountId + '}'
        method = build_api_gateway_method_name_reference(
            method_name=method_name)
        resource = build_api_gateway_resource_path_ref(
            resource_name=resource_name)
        return f'arn:aws:execute-api:{region}:{account_id}:{rest_api_id}/*/{method}{resource}'

    def _get_lambda_function_uri(self, lambda_name):
        region = self.config.region
        lambda_arn = build_function_invoke_arn_ref(lambda_name)
        arn = f'arn:aws:apigateway:{region}:lambda:path/2015-03-31/functions/{lambda_arn}/invocations'
        return arn

    def create_request_validator(self, method_meta, resource_name, api_name,
                                 http_method):
        request_validator_meta = method_meta.get('request_validator')
        if request_validator_meta:
            validator_params = \
                ApiGatewayConnection.get_request_validator_params(
                    request_validator_meta)
            validator_name = validator_params.get(REQ_VALIDATOR_PARAM_NAME)
            validate_body = validator_params.get(
                REQ_VALIDATOR_PARAM_VALIDATE_BODY)
            validate_param = validator_params.get(
                REQ_VALIDATOR_PARAM_VALIDATE_PARAMS)

            tf_validator_res_name = build_terraform_resource_name(api_name,
                                                                  resource_name,
                                                                  http_method,
                                                                  'Validator')
            request_validator = aws_api_gateway_request_validator(
                resource_name=tf_validator_res_name,
                validator_name=validator_name,
                validate_request_body=validate_body,
                validate_request_parameters=validate_param,
                rest_api_name=api_name)
            self.template.add_aws_api_gateway_request_validator(
                meta=request_validator)


def aws_api_gateway_request_validator(resource_name, validator_name,
                                      rest_api_name,
                                      validate_request_body=None,
                                      validate_request_parameters=None):
    validator = {
        'name': validator_name,
        'rest_api_id': build_rest_api_id_ref(api_name=rest_api_name)
    }
    if validate_request_body:
        validator['validate_request_body'] = validate_request_body
    if validate_request_parameters:
        validator['validate_request_parameters'] = validate_request_parameters
    resource = {
        resource_name: validator
    }
    return resource


def generate_tf_template_for_api_gateway(api_name):
    resource = {
        api_name:
            {
                "endpoint_configuration": [
                    {
                        "types": [
                            "REGIONAL"
                        ]
                    }
                ],
                "name": api_name
            }
    }
    return resource


def get_api_gateway_resource(resource_name, path_part, rest_api, parent_id):
    rest_api_id = build_rest_api_id_ref(api_name=rest_api)
    resource = {
        resource_name:
            {
                "parent_id": parent_id,
                "path_part": path_part,
                "rest_api_id": rest_api_id
            }
    }
    return resource


def api_gateway_method_response(method_name, status_code, http_method,
                                resource_name, api_name):
    response_name = build_terraform_resource_name(resource_name, http_method,
                                                  status_code)
    method = build_api_gateway_method_name_reference(method_name=method_name)
    resource_id = build_api_gateway_resource_id_ref(
        resource_name=resource_name)
    rest_api_id = build_rest_api_id_ref(api_name=api_name)
    resource = {
        response_name:
            {
                "http_method": method,
                "resource_id": resource_id,
                "rest_api_id": rest_api_id,
                "status_code": status_code
            }
    }
    return resource


def create_api_gateway_integration(integration_name, api_name,
                                   resource_name,
                                   method_name, integration_type,
                                   request_template=None,
                                   integration_method=None,
                                   uri=None,
                                   passthrough_behavior=None,
                                   credentials=None,
                                   cache_key_parameters=None):
    resource_id = build_api_gateway_resource_id_ref(
        resource_name=resource_name)
    rest_api_id = build_rest_api_id_ref(api_name=api_name)
    http_method = build_api_gateway_method_name_reference(method_name)

    integration = {
        "http_method": http_method,
        "resource_id": resource_id,
        "rest_api_id": rest_api_id,
        "type": integration_type,
        "uri": uri
    }

    if request_template:
        integration.update({"request_templates": request_template})

    if integration_method:
        integration.update({'integration_http_method': integration_method})

    if passthrough_behavior:
        integration.update({'passthrough_behavior': passthrough_behavior})

    if credentials:
        integration.update({'credentials': credentials})

    if cache_key_parameters:
        integration.update({'cache_key_parameters': cache_key_parameters})

    resource = {
        integration_name: integration
    }
    return resource


def api_gateway_stage(api_name, stage_name, deployment_name,
                      cache_cluster_configuration=None):
    rest_api_id = build_rest_api_id_ref(api_name=api_name)
    deployment_id = build_api_gateway_deployment_id_ref(
        deployment_name=deployment_name)

    stage_meta = {
        "deployment_id": deployment_id,
        "rest_api_id": rest_api_id,
        "stage_name": stage_name
    }

    if cache_cluster_configuration:
        cache_cluster_enabled = cache_cluster_configuration.get(
            'cache_enabled')
        cache_size_value = cache_cluster_configuration.get(
            'cache_size')
        cache_cluster_size = str(cache_size_value)

        if str(cache_cluster_enabled).lower() == 'true':
            stage_meta['cache_cluster_enabled'] = cache_cluster_enabled
        if cache_cluster_size:
            stage_meta['cache_cluster_size'] = cache_cluster_size

    stage_resource_name = build_terraform_resource_name(api_name, 'stage')
    resource = {
        stage_resource_name: stage_meta
    }
    return resource


def api_gateway_deployment(api_name, deployment_name, methods,
                           integration_names):
    dependencies = []
    for method in methods:
        dependencies.append(f'aws_api_gateway_method.{method}')
    for integration in integration_names:
        dependencies.append(f'aws_api_gateway_integration.{integration}')

    rest_api_id = build_rest_api_id_ref(api_name=api_name)
    resource = {
        deployment_name:
            {
                "rest_api_id": rest_api_id,
                "depends_on": dependencies
            }
    }
    return resource


def create_api_gateway_integration_response(resource_name, api_name,
                                            status_code, response_template,
                                            http_method, method_name,
                                            integration,
                                            selection_pattern=None):
    integration_response_name = build_terraform_resource_name(resource_name,
                                                              http_method,
                                                              status_code)
    method = build_api_gateway_method_name_reference(method_name)
    resource_id = build_api_gateway_resource_id_ref(
        resource_name=resource_name)
    rest_api_id = build_rest_api_id_ref(api_name=api_name)
    resource = {
        integration_response_name: [
            {
                "http_method": method,
                "resource_id": resource_id,
                "response_templates": response_template,
                "rest_api_id": rest_api_id,
                "status_code": status_code,
                "selection_pattern": selection_pattern,
                "depends_on": [
                    f'aws_api_gateway_integration.{integration}']
            }
        ]
    }
    return resource


def get_api_gateway_method(http_method, resource_name, rest_api,
                           method_name,
                           authorization=None,
                           authorizer_id=None,
                           request_parameters=None):
    resource_id = build_api_gateway_resource_id_ref(
        resource_name=resource_name)
    rest_api_id = build_rest_api_id_ref(api_name=rest_api)

    method_meta = {
        "authorization": authorization,
        "http_method": http_method,
        "resource_id": resource_id,
        "rest_api_id": rest_api_id
    }

    if authorizer_id:
        method_meta.update({'authorizer_id': authorizer_id})

    if request_parameters:
        method_meta.update({'request_parameters': request_parameters})

    resource = {
        method_name: [method_meta]
    }
    return resource


def get_api_gateway_stage(stage_name, deployment, rest_api):
    deployment_id = build_api_gateway_deployment_id_ref(
        deployment_name=deployment)
    rest_api_id = build_rest_api_id_ref(api_name=rest_api)
    resource = {
        stage_name: [
            {
                "deployment_id": deployment_id,
                "rest_api_id": rest_api_id,
                "stage_name": stage_name
            }
        ]
    }
    return resource


def aws_lambda_permissions(resource_name, function_name, api_source_arn):
    resource = {
        resource_name: [
            {
                "action": "lambda:InvokeFunction",
                "function_name": build_function_name_ref(function_name),
                "principal": "apigateway.amazonaws.com",
                "source_arn": api_source_arn,
                "statement_id": "AllowExecutionFromAPIGateway"
            }
        ]
    }
    return resource


def api_gateway_authorizer(authorizer_name,
                           authorizer_uri,
                           rest_api_name,
                           identity_source=None,
                           ttl=None,
                           auth_type=None):
    rest_api_id = build_rest_api_id_ref(api_name=rest_api_name)

    authorizer = {
        "authorizer_uri": authorizer_uri,
        "name": authorizer_name,
        "rest_api_id": rest_api_id
    }

    if identity_source:
        authorizer.update({'identity_source': identity_source})
    if ttl:
        authorizer.update({'authorizer_result_ttl_in_seconds': ttl})
    if auth_type:
        authorizer.update({'type': auth_type})

    resource = {
        authorizer_name: [authorizer]
    }
    return resource



import json
import os
from functools import partial

import click

from syndicate.core.constants import S3_BUCKET_ACL_LIST, \
    API_GW_AUTHORIZER_TYPES, CUSTOM_AUTHORIZER_KEY, \
    EC2_LAUNCH_TEMPLATE_SUPPORTED_IMDS_VERSIONS
from syndicate.core.generators.deployment_resources import *
from syndicate.core.generators.deployment_resources.api_gateway_generator import \
    ApiGatewayAuthorizerGenerator
from syndicate.core.generators.deployment_resources.ec2_launch_template_generator import \
    EC2LaunchTemplateGenerator
from syndicate.core.generators.lambda_function import PROJECT_PATH_PARAM
from syndicate.core.helper import OrderedGroup, OptionRequiredIf, \
    validate_incompatible_options, validate_authorizer_name_option, \
    verbose_option, validate_api_gw_path
from syndicate.core.helper import ValidRegionParamType
from syndicate.core.helper import check_bundle_bucket_name
from syndicate.core.helper import resolve_project_path, timeit

GENERATE_META_GROUP_NAME = 'meta'
dynamodb_type_param = click.Choice(['S', 'N', 'B'])


@click.group(name=GENERATE_META_GROUP_NAME, cls=OrderedGroup)
@click.option('--project_path', nargs=1,
              help="Path to the project folder. Default value: the one "
                   "from the current config if it exists. "
                   "Otherwise - the current working directory",
              callback=resolve_project_path)
@click.pass_context
def meta(ctx, project_path):
    """Generates deployment resources templates"""
    if not os.access(project_path, os.F_OK):
        click.echo(f"The provided path {project_path} doesn't exist")
        return
    elif not os.access(project_path, os.W_OK) or not os.access(project_path,
                                                               os.X_OK):
        click.echo(f"Incorrect permissions for the provided path "
                   f"'{project_path}'")
        return
    ctx.ensure_object(dict)
    ctx.obj[PROJECT_PATH_PARAM] = project_path


@meta.command(name='dax_cluster')
@click.option('--resource_name', required=True, type=str,
              help="Dax cluster name")
@click.option('--node_type', required=True, type=str,
              help="The node type for the nodes in the cluster")
@click.option('--iam_role_name', required=True, type=str,
              help="Role name to access DynamoDB tables")
@click.option('--subnet_group_name', required=True, type=str,
              help='The name of the subnet group to be used for the '
                   'replication group')
@click.option('--subnet_ids', type=str, multiple=True,
              help='Subnet ids to create a subnet group from. Don\'t specify '
                   'in case of using existing subnet group')
@click.option('--cluster_endpoint_encryption_type',
              type=click.Choice(['NONE', 'TLS']), default='TLS',
              help='The encryption type of the cluster\'s endpoint. '
                   'The default value is \'TLS\'')
@click.option('--parameter_group_name', type=str,
              help='The parameter group to be associated with the DAX cluster')
@verbose_option
@click.pass_context
@timeit()
def dax_cluster(ctx, **kwargs):
    """Generated dax cluster deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = DaxClusterGenerator(**kwargs)
    _generate(generator)
    click.echo(f'Dax cluster \'{kwargs["resource_name"]}\' was '
               f'successfully generated')


@meta.command(name='dynamodb')
@click.option('--resource_name', required=True, type=str,
              help="DynamoDB table name")
@click.option('--hash_key_name', required=True, type=str,
              help="DynamoDB table hash key")
@click.option('--hash_key_type', required=True,
              type=click.Choice(['S', 'N', 'B']),
              help="DynamoDB hash key type")
@click.option('--sort_key_name', type=str,
              help="DynamoDB sort key. If not specified, the table will have "
                   "only a hash key")
@click.option('--sort_key_type', type=dynamodb_type_param,
              cls=OptionRequiredIf, required_if='sort_key_name',
              help="Required if sort key name is specified")
@click.option('--read_capacity', type=int,
              help="The maximum number of strongly consistent reads that can"
                   "be performed per second. If not specified, sets the "
                   "default value to 1")
@click.option('--write_capacity', type=int,
              help="The maximum number of writing processes consumed per"
                   "second. If not specified, sets the default value to 1")
@verbose_option
@click.pass_context
@timeit()
def dynamodb(ctx, **kwargs):
    """Generates dynamoDB deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = DynamoDBGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Table '{kwargs['resource_name']}' was added successfully!")


@meta.command(name='dynamodb_global_index')
@click.option('--table_name', required=True, type=str,
              help="DynamoDB table name to add index to")
@click.option('--name', required=True, type=str,
              help="Index name")
@click.option('--index_key_name', required=True, type=str,
              help="Index hash key")
@click.option('--index_key_type', required=True,
              type=dynamodb_type_param,
              help='Hash key index type')
@click.option('--index_sort_key_name', type=str,
              help='Index sort key')
@click.option('--index_sort_key_type', type=dynamodb_type_param,
              cls=OptionRequiredIf, required_if='index_sort_key_name',
              help="Sort key type")
@verbose_option
@click.pass_context
@timeit()
def dynamodb_global_index(ctx, **kwargs):
    """Adds dynamodb global index to existing dynamodb table"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = DynamoDBGlobalIndexGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Global index '{kwargs['name']}' was added successfully")


@meta.command(name='dynamodb_autoscaling')
@click.option('--table_name', type=str, required=True,
              help="DynamoDB table name to add autoscaling to")
@click.option('--policy_name', type=str, required=True,
              help="Autoscaling policy name")
@click.option('--min_capacity', type=click.IntRange(min=1),
              help="Minimum capacity level. If not specified, sets the default"
                   " value to 1")
@click.option('--max_capacity', type=click.IntRange(min=1),
              help="Maximum capacity level. If not specified, sets the default"
                   " value to 10")
@click.option('--target_utilization', type=click.IntRange(min=20, max=90),
              help="Target utilization in autoscaling. If not specified, sets "
                   "the default value to 70 %")
@click.option('--scale_in_cooldown', type=click.IntRange(min=0),
              help="Scaling policy value of in cooldown in seconds. Is not "
                   "specified, sets the default value to 60")
@click.option('--scale_out_cooldown', type=click.IntRange(min=0),
              help="Scaling policy value of out cooldown in seconds. Is not "
                   "specified, sets the default value to 60")
@click.option('--dimension', type=str,
              help="Autoscaling dimension. If not specified, sets the default"
                   "the default value to 'dynamodb:table:ReadCapacityUnits'")
@click.option('--role_name', type=str,
              help="The name of the role, which performs autoscaling. If not "
                   "specified, sets the value to default service linked role: "
                   "'AWSServiceRoleForApplicationAutoScaling_DynamoDBTable'")
@verbose_option
@click.pass_context
@timeit()
def dynamodb_autoscaling(ctx, **kwargs):
    """Adds autoscaling settings to existing dynamodb table"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = DynamoDBAutoscalingGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Autoscaling setting to table '{kwargs['table_name']}' was "
               f"added successfully")


@meta.command(name='s3_bucket')
@click.option('--resource_name', required=True, type=str,
              help="S3 bucket name", callback=check_bundle_bucket_name)
@click.option('--location', type=ValidRegionParamType(),
              help="The region where the bucket is created, the default value "
                   "is the region set in syndicate config")
@click.option('--acl', type=click.Choice(S3_BUCKET_ACL_LIST),
              help="The channel ACL to be applied to the bucket. If not "
                   "specified, sets the default value to 'private'")
@click.option('--block_public_acls', type=bool, required=False,
              is_eager=True,
              help='Specifies whether Amazon S3 should block public access '
                   'control lists (ACLs) for this bucket and objects in this '
                   'bucket. Default value is True')
@click.option('--ignore_public_acls', type=bool, required=False,
              is_eager=True,
              help='Specifies whether Amazon S3 should ignore public ACLs for '
                   'this bucket and objects in this bucket. Default value '
                   'is True')
@click.option('--block_public_policy', type=bool, required=False,
              is_eager=True,
              help='Specifies whether Amazon S3 should block public bucket '
                   'policies for this bucket. Default value is True')
@click.option('--restrict_public_buckets', type=bool, required=False,
              is_eager=True,
              help='Specifies whether Amazon S3 should restrict public bucket '
                   'policies for this bucket. Default value is True')
@click.option('--static_website_hosting', type=bool, required=False,
              callback=partial(validate_incompatible_options,
                               incompatible_options=['block_public_acls',
                                                     'ignore_public_acls',
                                                     'restrict_public_buckets',
                                                     'block_public_policy']),
              help='Specifies whether the S3 bucket should be configured for '
                   'static WEB site hosting. If specified public read access '
                   'will be configured for all S3 bucket objects! Default '
                   'value is False')
@verbose_option
@click.pass_context
@timeit()
def s3_bucket(ctx, **kwargs):
    """Generates s3 bucket deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = S3Generator(**kwargs)
    _generate(generator)
    click.echo(f"S3 bucket '{kwargs['resource_name']}' was "
               f"added successfully!")


@meta.command(name='api_gateway')
@click.option('--resource_name', required=True, type=str,
              help="Api gateway name")
@click.option('--deploy_stage', required=True, type=str,
              help="The stage to deploy the API")
@click.option('--minimum_compression_size',
              type=click.IntRange(min=0, max=10 * 1024 * 1024),
              help="Compression size for api gateway. If not specified, "
                   "compression will be disabled")
@verbose_option
@click.pass_context
@timeit()
def api_gateway(ctx, **kwargs):
    """Generates api gateway deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = ApiGatewayGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Api gateway '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name='web_socket_api_gateway')
@click.option('--resource_name', required=True, type=str,
              help="Api gateway name")
@click.option('--deploy_stage', required=True, type=str,
              help="The stage to deploy the API")
@verbose_option
@click.pass_context
@timeit()
def web_socket_api_gateway(ctx, **kwargs):
    """Generates web socket api gateway deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = WebSocketApiGatewayGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Api gateway '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name='api_gateway_authorizer')
@click.option('--api_name', required=True, type=str,
              help="Api gateway name to add index to")
@click.option('--name', required=True, type=str,
              help="Authorizer name")
@click.option('--type', type=click.Choice(API_GW_AUTHORIZER_TYPES),
              required=True, help="Authorizer type. 'TOKEN' for a Lambda "
                                  "function using a single authorization "
                                  "token submitted in a custom header, "
                                  "'REQUEST' for a Lambda function using "
                                  "incoming request parameters, and "
                                  "'COGNITO_USER_POOLS' for using an Amazon "
                                  "Cognito user pool")
@click.option('--provider_name', type=str, required=True,
              help="Identity provider name")
@verbose_option
@click.pass_context
@timeit()
def api_gateway_authorizer(ctx, **kwargs):
    """Adds authorizer to an existing api gateway"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = ApiGatewayAuthorizerGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Authorizer '{kwargs['name']}' was added to API gateway "
               f"'{kwargs['api_name']}' successfully")


@meta.command(name='api_gateway_resource')
@click.option('--api_name', required=True, type=str,
              help="Api gateway name to add index to")
@click.option('--path', required=True, callback=validate_api_gw_path,
              help="Resource path to create")
@click.option('--enable_cors', type=bool,
              help="Enables CORS on the resourcemethod. If not specified, sets"
                   "the default value to False")
@verbose_option
@click.pass_context
@timeit()
def api_gateway_resource(ctx, **kwargs):
    """Adds resource to existing api gateway"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = ApiGatewayResourceGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Resource '{kwargs['path']}' was added to API gateway "
               f"'{kwargs['api_name']}' successfully")


@meta.command(name='api_gateway_resource_method')
@click.option('--api_name', required=True, type=str,
              help="Api gateway name to add index to")
@click.option('--path', required=True, callback=validate_api_gw_path,
              help="Resource path to create")
@click.option('--method', required=True,
              type=click.Choice(['POST', 'GET', 'DELETE', 'PUT', 'HEAD',
                                 'PATCH', 'ANY']),
              help="Resource method to add")
@click.option('--integration_type', type=str,
              help="The resource which the method is connected to: "
                   "[lambda|service|http|mock]. If not specified, sets the "
                   "default value to 'mock'")
@click.option('--lambda_name', type=str, help="Lambda name. Required if "
                                              "integration type is lambda")
@click.option('--lambda_region', type=ValidRegionParamType(),
              help="The region where the lambda is located. If not specified, "
                   "sets the default value from syndicate config")
@click.option('--authorization_type', is_eager=True,
              type=click.Choice(["NONE", "AWS_IAM", CUSTOM_AUTHORIZER_KEY]),
              help="The method's authorization type. If not specified, sets "
                   "the default value to 'NONE'")
@click.option('--authorizer_name', type=str,
              callback=validate_authorizer_name_option,
              help="The method's authorizer name can be used only with "
                   "'--authorization_type' 'CUSTOM'")
@click.option('--api_key_required', type=bool,
              help="Specifies whether the method requires a valid API key. "
                   "If not specified, the default value is set to False")
@verbose_option
@click.pass_context
@timeit()
def api_gateway_resource_method(ctx, **kwargs):
    """Adds a method to existing api gateway resource"""

    if kwargs.get('integration_type') == 'lambda' \
            and not kwargs.get('lambda_name'):
        raise click.MissingParameter(
            "Lambda name is required if the integration type is 'lambda'",
            param_type='option', param_hint='lambda_name')

    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = ApiGatewayResourceMethodGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Method '{kwargs['method']}' was added to API gateway "
               f"resource '{kwargs['path']}' successfully")


@meta.command(name='iam_policy')
@click.option('--resource_name', required=True, type=str,
              help='IAM policy name')
@click.option('--policy_content', help='The path to JSON file with IAM policy '
                                       'content. If not specified, template '
                                       'value will be set',
              type=click.File(mode='r'))
@verbose_option
@click.pass_context
@timeit()
def iam_policy(ctx, **kwargs):
    """Generates IAM policy deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    if kwargs['policy_content']:
        try:
            kwargs['policy_content'] = json.load(kwargs['policy_content'])
        except json.decoder.JSONDecodeError as e:
            raise click.BadParameter(str(e), param_hint='policy_content')
    generator = IAMPolicyGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Iam policy '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name='iam_role')
@click.option('--resource_name', required=True, type=str,
              help="IAM role name")
@click.option('--principal_service', required=True, type=str,
              help="The service which will use the role")
@click.option("--predefined_policies", type=str, multiple=True,
              help="Managed IAM policies list")
@click.option("--custom_policies", type=str, multiple=True,
              help="Customer AWS policies names")
@click.option("--allowed_accounts", type=str, multiple=True,
              help="The list of accounts, which can assume the role")
@click.option("--external_id", type=str, help="External ID in role")
@click.option("--instance_profile", type=bool,
              help="If true, instance profile with role name is created")
@click.option('--permissions_boundary', type=str,
              help="The name or the ARN of permissions boundary policy to "
                   "attach to this role")
@verbose_option
@click.pass_context
@timeit()
def iam_role(ctx, **kwargs):
    """Generates IAM role deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = IAMRoleGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Iam role '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name='kinesis_stream')
@click.option('--resource_name', type=str, required=True,
              help="Kinesis stream name")
@click.option('--shard_count', type=int, required=True,
              help="Number of shards that the stream uses")
@verbose_option
@click.pass_context
@timeit()
def kinesis_stream(ctx, **kwargs):
    """Generates kinesis stream deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = KinesisStreamGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Kinesis stream '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name='sns_topic')
@click.option('--resource_name', type=str, required=True,
              help="SNS topic name")
@click.option('--region', type=ValidRegionParamType(allowed_all=True),
              required=True, help="Where the topic should be deployed")
@verbose_option
@click.pass_context
@timeit()
def sns_topic(ctx, **kwargs):
    """Generates sns topic deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = SNSTopicGenerator(**kwargs)
    _generate(generator)
    click.echo(f"SNS topic '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name='step_function')
@click.option('--resource_name', type=str, required=True,
              help="Step function name")
@click.option('--iam_role', type=str, required=True,
              help="IAM role to use for this state machine")
@verbose_option
@click.pass_context
@timeit()
def step_function(ctx, **kwargs):
    """Generate step function deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = StepFunctionGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Step function '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name='step_function_activity')
@click.option('--resource_name', type=str, required=True,
              help="Step function activity name")
@verbose_option
@click.pass_context
@timeit()
def step_function_activity(ctx, **kwargs):
    """Generates step function activity deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = StepFunctionActivityGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Step function activity '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name='ec2_instance')
@click.option('--resource_name', type=str, required=True,
              help="Instance name")
@click.option('--key_name', type=str, required=True,
              help="SHH key to access the instance")
@click.option('--image_id', type=str, required=True,
              help="Image id to create the instance from")
@click.option('--instance_type', type=str,
              help="Instance type")
@click.option('--disable_api_termination', type=bool,
              help="Api termination protection. Default value is True")
@click.option('--security_group_ids', type=str, multiple=True,
              help="Security group ids")
@click.option('--security_group_names', type=str, multiple=True,
              help="Security group names")
@click.option('--availability_zone', type=str,
              help="Instance availability zone")
@click.option('--subnet_id', type=str, cls=OptionRequiredIf,
              required_if="availability_zone",
              help="Subnet ID (required if availability zone is set)")
@click.option('--userdata_file', type=str,
              help="File path to userdata (file relative pathname from the"
                   "directory which is set up in the env variable 'SDCT_CONF'")
@click.option('--iam_role', type=str,
              help="Instance IAM role")
@verbose_option
@click.pass_context
@timeit()
def ec2_instance(ctx, **kwargs):
    """Generates ec2 instance deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = EC2InstanceGenerator(**kwargs)
    _generate(generator)
    click.echo(f"EC2 instance '{kwargs['resource_name']}' was added"
               f"successfully")


@meta.command(name='ec2_launch_template')
@click.option('--resource_name', type=str, required=True,
              help="Launch template name")
@click.option('--image_id', type=str, required=True,
              help="The ID of the AMI")
@click.option('--key_name', type=str,
              help="The name of the key pair")
@click.option('--instance_type', type=str,
              help="Instance type")
@click.option('--security_group_ids', type=str, multiple=True,
              help="Security group ids")
@click.option('--security_group_names', type=str, multiple=True,
              help="Security group names")
@click.option('--userdata_file', type=str,
              help="File path to userdata (can be specified as a relative "
                   "path to the project path)")
@click.option('--iam_role', type=str,
              help="Instance IAM role")
@click.option('--imds_version', type=click.Choice(
    EC2_LAUNCH_TEMPLATE_SUPPORTED_IMDS_VERSIONS),
              help="IMDS version")
@click.option('--version_description', type=str,
              help="A description for the version of the launch template")
@verbose_option
@click.pass_context
@timeit()
def ec2_launch_template(ctx, **kwargs):
    """Generates ec2_launch_template deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = EC2LaunchTemplateGenerator(**kwargs)
    _generate(generator)
    click.echo(f"ec2_launch_template '{kwargs['resource_name']}' was added "
               f"successfully")


@meta.command(name='sqs_queue')
@click.option('--resource_name', type=str, required=True,
              help="SQS queue name")
@click.option('--region', type=ValidRegionParamType(),
              help="The region where the queue is deployed. Default value is "
                   "the one from syndicate config")
@click.option('--fifo_queue', type=bool,
              help="If True, the queue is FIFO. Default value is False")
@click.option('--visibility_timeout', type=click.IntRange(min=0, max=43200),
              help="The visibility timeout for the queue. Default value is 30")
@click.option('--delay_seconds', type=click.IntRange(min=0, max=900),
              help="The length of time in seconds for which the delivery "
                   "of all the messages in the queue is delayed. Default "
                   "value is 0")
@click.option('--maximum_message_size',
              type=click.IntRange(min=1024, max=262144),
              help="The limit of how many bytes a message can contain before "
                   "Amazon SQS rejects it. Default value is 1024")
@click.option('--message_retention_period', type=click.IntRange(min=60,
                                                                max=1209600),
              help="The length of time in seconds for which Amazon SQS "
                   "retains a message. Default value is 60")
@click.option('--receive_message_wait_time_seconds',
              type=click.IntRange(min=0, max=20),
              help="The length of time in seconds for which a 'ReceiveMessage'"
                   " action waits for a message to arrive")
@click.option('--dead_letter_target_arn', type=str,
              help="Arn of a dead-letter queue Amazon SQS moves messages "
                   "after the value of maxReceiveCount is exceeded")
@click.option('--max_receive_count', type=click.IntRange(min=1, max=1000),
              help="The number of times a message is delivered to the source "
                   "queue before being moved to the dead-letter queue. "
                   "Required if 'dead_letter_target_arn' is specified",
              cls=OptionRequiredIf, required_if='dead_letter_target_arn')
@click.option('--kms_master_key_id', type=str,
              help="The id of an AWS-managed customer master key (CMK) for "
                   "Amazon SQS or a custom CMK")
@click.option('--kms_data_key_reuse_period_seconds',
              type=click.IntRange(min=60, max=86400),
              help="The length of time in seconds for which Amazon SQS can "
                   "reuse a data key to encrypt or decrypt messages before "
                   "calling AWS KMS again")
@click.option('--content_based_deduplication', type=bool,
              help="Enables content-based deduplication")
@verbose_option
@click.pass_context
@timeit()
def sqs_queue(ctx, **kwargs):
    """Generates sqs queue deployment deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = SQSQueueGenerator(**kwargs)
    _generate(generator)
    click.echo(f"SQS queue '{kwargs['resource_name']}' was added "
               f"successfully")


@meta.command(name="sns_application")
@click.option('--resource_name', type=str, required=True,
              help="The name of the sns application")
@click.option('--platform', required=True,
              type=click.Choice(['GCM', 'ADM', 'APNS', 'APNS_SANDBOX']),
              help="SNS application platform")
@click.option('--region', type=ValidRegionParamType(),
              help="The region where the application is deployed. Default "
                   "value is the one from syndicate config")
@click.option('--attributes', type=(str, str), multiple=True,
              help="SNS application attributes")
@verbose_option
@click.pass_context
@timeit()
def sns_application(ctx, **kwargs):
    """Generates sns application deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = SNSApplicationGenerator(**kwargs)
    _generate(generator)
    click.echo(f"SNS application '{kwargs['resource_name']}' was added "
               f"successfully")


@meta.command(name="cognito_user_pool")
@click.option('--resource_name', type=str, required=True,
              help="Cognito user pool name")
# @click.option('--region', type=ValidRegionParamType(), required=True,
#               help="The region where the user pool is created")
@click.option('--auto_verified_attributes',
              type=click.Choice(['phone_number', 'email']),
              help="The attributes to be auto-verified. "
                   "Default value is email", multiple=True)
@click.option('--sns_caller_arn', type=str,
              help="The arn of the IAM role in your account which Cognito "
                   "will use to send SMS messages. Required if 'phone_number' "
                   "in 'auto_verified_attributes' is specified")
@click.option('--username_attributes',
              type=click.Choice(['phone_number', 'email']),
              help="Specifies whether email addresses or phone numbers can "
                   "be specified as usernames when a user signs up. Default "
                   "value is email", multiple=True)
@click.option('--custom_attributes', type=(str, str), multiple=True,
              help="A list of custom attributes: (name type)")
@verbose_option
@click.pass_context
@timeit()
def cognito_user_pool(ctx, **kwargs):
    """Generates cognito user pool deployment resource template"""
    if 'phone_number' in kwargs['auto_verified_attributes'] \
            and not kwargs.get('sns_caller_arn'):
        raise click.MissingParameter("Sns caller IAM role arn is required when"
                                     " 'phone_number' is specified in "
                                     "'auto_verified_attributes'",
                                     param_type='option',
                                     param_hint='sns_caller_arn')

    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = CognitoUserPoolGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Cognito user pool '{kwargs['resource_name']}' was added "
               f"successfully")


@meta.command(name="cognito_federated_pool")
@click.option('--resource_name', type=str, required=True,
              help="Cognito federated pool name")
@click.option('--auth_role', type=str,
              help="IAM role for authorized users")
@click.option('--unauth_role', type=str,
              help="IAM role for unauthorized users")
@click.option('--open_id_providers', type=str, multiple=True,
              help="A list of OpenID Connect providers")
@click.option('--provider_name', type=str,
              help="Developer provider name")
@verbose_option
@click.pass_context
@timeit()
def cognito_federated_pool(ctx, **kwargs):
    """Generates cognito federated pool deployment resource template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = CognitoFederatedPoolGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Cognito federated pool '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name='batch_compenv')
@click.option('--resource_name', type=str, required=True,
              help="Batch compute environment name")
@click.option('--compute_environment_type',
              type=click.Choice(['MANAGED', 'UNMANAGED']),
              help="The type of compute environment. "
                   "Default value is 'MANAGED'")
@click.option('--allocation_strategy',
              type=click.Choice(['BEST_FIT', 'BEST_FIT_PROGRESSIVE',
                                 'SPOT_CAPACITY_OPTIMIZED']),
              help="The allocation strategy to use for the compute resource "
                   "if not enough instances of the best fitting instance type "
                   "can be allocated")
@click.option('--state', type=click.Choice(['ENABLED', 'DISABLED']),
              help="The state of compute environment")
@click.option('--service_role', type=str,
              help="The full Amazon Resource Name (ARN) of the IAM role that "
                   "allows Batch to make calls to other Amazon Web Services "
                   "services on your behalf. If not specified, role "
                   "'AWSBatchServiceRole' will be taken if it exists, if "
                   "doesn't it'll be created")
@click.option('--type',
              type=click.Choice(['EC2', 'SPOT', 'FARGATE', 'FARGATE_SPOT']),
              help="The type of compute environment. Default value is EC2")
@click.option('--minv_cpus', type=click.IntRange(min=0),
              help='The minimum number of Amazon EC2 vCPUs that an '
                   'environment should maintain. Default value is 0')
@click.option('--maxv_cpus', type=click.IntRange(min=1),
              help="The maximum number of Amazon EC2 vCPUs that a compute "
                   "environment can reach. Default value is 8")
@click.option('--desiredv_cpus', type=int,
              help="The desired number of Amazon EC2 vCPUS in the compute "
                   "environment. Default value is 1")
@click.option('--instance_types', type=str, multiple=True,
              help="The instances types that can be launched. Default value "
                   "is 'optimal'")
@click.option('--security_group_ids', type=str, multiple=True, required=True,
              help="The Amazon EC2 security groups associated with instances "
                   "launched in the compute environment")
@click.option('--subnets', type=str, multiple=True, required=True,
              help="The VPC subnets where the compute resources are launched")
@click.option('--instance_role', type=str,
              help="The Amazon ECS instance profile applied to Amazon EC2 "
                   "instances in a compute environment")
@verbose_option
@click.pass_context
@timeit()
def batch_compenv(ctx, **kwargs):
    """Generates batch compenv deployment resources template"""
    if kwargs.get('type') != 'FARGATE':
        if not kwargs.get('instance_role'):
            raise click.MissingParameter("'instance_role' is required if "
                                         "batch compenv type ISN'T 'FARGATE'",
                                         param_type='option',
                                         param_hint='instance_role')
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = BatchCompenvGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Batch compute environment '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name='batch_jobdef')
@click.option('--resource_name', type=str, required=True,
              help='Batch job definition name')
@click.option('--job_definition_type', required=True,
              type=click.Choice(['container', 'multinode']),
              help='The type of job definition')
@click.option('--image', type=str,
              help='The image used to start a container. '
                   'Default value is \'alpine\'')
@click.option('--job_role_arn', type=str,
              help='The ARN of the IAM role that the container can assume for '
                   'AWS permissions')
@verbose_option
@click.pass_context
@timeit()
def batch_jobdef(ctx, **kwargs):
    """Generates batch job definition deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = BatchJobdefGenerator(**kwargs)
    _generate(generator)
    click.echo(f'Batch job definition \'{kwargs["resource_name"]}\' was '
               f'added successfully')


@meta.command(name="batch_jobqueue")
@click.option('--resource_name', type=str, required=True,
              help="Batch job queue name")
@click.option('--state', type=click.Choice(["ENABLED", "DISABLED"]),
              help="The state of the job queue. Default value is 'ENABLED'")
@click.option('--priority', type=int, help="The priority of the job queue. "
                                           "Default value is 1")
@click.option('--compute_environment_order', type=(int, str), multiple=True,
              help="The set of compute environments mapped to a job queue and "
                   "their order relative to each other. (order, compute_env)")
@verbose_option
@click.pass_context
@timeit()
def batch_jobqueue(ctx, **kwargs):
    """Generates batch job queue deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = BatchJobqueueGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Batch job queue '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name="cloudwatch_alarm")
@click.option('--resource_name', type=str, required=True,
              help="Cloudwatch alarm name")
@click.option('--metric_name', type=str, required=True,
              help="The metric's name")
@click.option('--namespace', type=str, required=True,
              help="The namespace for the metric associated with the alarm")
@click.option('--description', type=str, help="The description for the alarm")
@click.option('--period', type=click.IntRange(min=1),
              help="The period in seconds over which the specified statistic "
                   "is applied. Valid values are 10, 30 and any multiple"
                   " of 60. Default value is 1200")
@click.option('--evaluation_periods', type=click.IntRange(min=1),
              help="The number of periods over which data is compared to the "
                   "specified threshold. Default value is 1")
@click.option('--threshold', type=float,
              help="The value to compare with the specified statistic. "
                   "Default value is 1.0")
@click.option('--comparison_operator',
              type=click.Choice(['GreaterThanOrEqualToThreshold',
                                 'GreaterThanThreshold',
                                 'LessThanThreshold',
                                 'LessThanOrEqualToThreshold',
                                 'LessThanLowerOrGreaterThanUpperThreshold',
                                 'LessThanLowerThreshold',
                                 'GreaterThanUpperThreshold']),
              help="An arithmetic operator to use when comparing the specified"
                   " statistic and threshold. The specified statistic value is"
                   " used as the first operand. Default value is "
                   "'GreaterThanOrEqualToThreshold'")
@click.option('--statistic', type=click.Choice(["SampleCount", "Average",
                                                'Sum', 'Minimum', 'Maximum']),
              help="The statistic for the metric associated with the alarm,"
                   "other than percentile. For percentile statistic use "
                   "'ExtendedStatistic'. Default value is 'SampleCount'")
@click.option('--sns_topics', type=str, multiple=True,
              help="The sns topics to execute when the alarm goes to an ALARM "
                   "state from any other state")
@click.option('--lambdas', type=str, multiple=True,
              help="The lambdas to execute when the alarm goes to an ALARM "
                   "state from any other state. Use `:` after lambda name to "
                   "specify alias or version")
@click.option('--ssm_response_plan', type=str, multiple=True,
              help="The response plan name to execute when the alarm goes to "
                   "an ALARM state from any other state")
@click.option('--evaluate_low_sample_count_percentile',
              type=click.Choice(["evaluate", "ignore"]),
              help="Only for percentiles-based alarms. Use 'ignore' and the "
                   "alarm state remains unchanged during periods with "
                   "insufficient data points for statistical significance. If "
                   "'evaluate' is specified (or parameter is omitted), the "
                   "alarm is always assessed and "
                   "may change state regardless of data point availability")
@click.option('--datapoints', type=click.IntRange(min=1),
              help="The number of datapoints that must be breaching to "
                   "trigger the alarm")
@verbose_option
@click.pass_context
@timeit()
def cloudwatch_alarm(ctx, **kwargs):
    """Generates Cloudwatch alarm deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = CloudWatchAlarmGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Cloudwatch alarm '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name="cloudwatch_event_rule")
@click.option('--resource_name', type=str, required=True,
              help="Cloudwatch event rule name")
@click.option('--rule_type', required=True, help="Cloudwatch event rule type",
              type=click.Choice(['schedule', 'ec2', 'api_call']))
@click.option('--expression', type=str,
              help="Rule expression (cron schedule). Valuable only if "
                   "rule_type is 'schedule'")
@click.option('--aws_service', type=str,
              help="The name of AWS service which the rule listens to. "
                   "Required only if rule_type is 'api_call'")
@click.option('--region', type=ValidRegionParamType(allowed_all=True),
              help="The region where the rule is deployed. Default value is "
                   "the one from syndicate config")
@verbose_option
@click.pass_context
@timeit()
def cloudwatch_event_rule(ctx, **kwargs):
    """Generates Cloudwatch event rule deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = CloudwatchEventRuleGenerator(**kwargs)
    _generate(generator)
    click.echo(f"Cloudwatch event rule '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name="eventbridge_rule")
@click.option('--resource_name', type=str, required=True,
              help="EventBridge rule name")
@click.option('--rule_type', required=True, help="EventBridge rule type",
              type=click.Choice(['schedule', 'ec2', 'api_call']))
@click.option('--expression', type=str,
              help="Rule expression (cron schedule). Valuable only if "
                   "rule_type is 'schedule'")
@click.option('--aws_service', type=str,
              help="The name of AWS service which the rule listens to. "
                   "Required only if rule_type is 'api_call'")
@click.option('--region', type=ValidRegionParamType(allowed_all=True),
              help="The region where the rule is deployed. Default value is "
                   "the one from syndicate config")
@verbose_option
@click.pass_context
@timeit()
def eventbridge_rule(ctx, **kwargs):
    """Generates EventBridge rule deployment resources-template
    claiming compatibility with Cloudwatch event rule generator"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = EventBridgeRuleGenerator(**kwargs)
    _generate(generator)
    click.echo(f"EventBridge rule '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name="documentdb_cluster")
@click.option('--resource_name', type=str, required=True,
              help="DocumentDB cluster name")
@click.option('--master_username', type=str, required=True,
              help="DocumentDB login ID for the master user")
@click.option('--master_password', type=str, required=True,
              help="The password for master user")
@click.option('--port', type=int,
              help="The port number on which the instances in the cluster "
                   "accept connections. Default value is 27017")
@click.option('--vpc_security_group_ids', type=str, multiple=True,
              help="A list of EC2 VPC security groups to associate with this "
                   "cluster. Is not specified, default security group is used")
@click.option('--availability_zones', type=str, multiple=True,
              help="A list of Amazon EC2 Availability Zones that instances in "
                   "the cluster can be created in. "
                   "If not specified default is used")
@verbose_option
@click.pass_context
@timeit()
def documentdb_cluster(ctx, **kwargs):
    """Generates documentdb cluster deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = DocumentDBClusterGenerator(**kwargs)
    _generate(generator)
    click.echo(f"DocumentDB cluster '{kwargs['resource_name']}' was "
               f"added successfully")


@meta.command(name="documentdb_instance")
@click.option('--resource_name', type=str, required=True,
              help="DocumentDB instance name")
@click.option('--cluster_identifier', type=str, required=True,
              help="The identifier of the cluster that the instance will "
                   "belong to")
@click.option('--instance_class', type=str,
              help="The compute and memory capacity of the instance. Default "
                   "value is 'db.r5.large'")
@click.option('--availability_zone', type=str,
              help="The Amazon EC2 Availability Zone that the instance is "
                   "created in. If not specified a random zone it the "
                   "endpoint's region is set")
@verbose_option
@click.pass_context
@timeit()
def documentdb_instance(ctx, **kwargs):
    """Generates documentdb instance deployment resources template"""
    kwargs[PROJECT_PATH_PARAM] = ctx.obj[PROJECT_PATH_PARAM]
    generator = DocumentDBInstanceGenerator(**kwargs)
    _generate(generator)
    click.echo(f"DocumentDB instance '{kwargs['resource_name']}' was "
               f"added successfully")


def _generate(generator: BaseConfigurationGenerator):
    """Just some common actions for this module are gathered in here"""
    try:
        generator.write()
    except ValueError as e:
        raise click.BadParameter(e)
    except RuntimeError as e:
        raise click.Abort(e)
    except Exception as e:
        raise Exception(f"An unexpected error occurred: {e}")



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import KINESIS_STREAM_TYPE


class KinesisStreamGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = KINESIS_STREAM_TYPE
    CONFIGURATION = {
        'shard_count': None,
    }



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.core.constants import S3_PATH_NAME
from syndicate.core.resources.helper import validate_params
from syndicate.core.resources.lambda_resource import \
    LAMBDA_LAYER_REQUIRED_PARAMS
from syndicate.core.transform.terraform.converter.tf_resource_converter import \
    TerraformResourceConverter
from syndicate.core.transform.terraform.tf_resource_name_builder import \
    lambda_layer_name


class LambdaLayerConverter(TerraformResourceConverter):

    def convert(self, name, resource):
        validate_params(name, resource, LAMBDA_LAYER_REQUIRED_PARAMS)

        key = resource[S3_PATH_NAME]
        description = resource.get('description')
        s3_bucket = self.config.deploy_target_bucket
        runtimes = resource.get('runtimes')
        licenses = resource.get('license')

        resource_name = lambda_layer_name(layer_name=name)
        layer = aws_lambda_layer_version(resource_name=resource_name,
                                         layer_name=name,
                                         description=description,
                                         s3_bucket=s3_bucket, s3_key=key,
                                         runtimes=runtimes,
                                         license_info=licenses)
        self.template.add_aws_lambda_layer_version(meta=layer)


def aws_lambda_layer_version(resource_name, layer_name, description=None,
                             license_info=None, runtimes=None,
                             s3_key=None, s3_bucket=None):
    layer = {
        'layer_name': layer_name
    }

    if description:
        layer['description'] = description
    if license_info:
        layer['license_info'] = license_info
    if runtimes:
        layer['compatible_runtimes'] = runtimes
    if s3_key:
        layer['s3_key'] = s3_key
    if s3_bucket:
        layer['s3_bucket'] = s3_bucket

    resource = {
        resource_name: layer
    }
    return resource



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.core.build.validator.batch_compenv_validator import _validate_field_type

JOB_DEFINITION_TYPES = ('container', 'multinode')


def validate_batch_jobdef(jobdef_name, jobdef_meta):
    """
    Performs check of Batch Job Definition resources.
    :param jobdef_name: name of resource
    :param jobdef_meta: resource definition

    :raises AssertionError in case of invalidity.

    :return: None
    """
    jobdef_config = [
        {
            'field_name': 'job_definition_name',
            'field_value': jobdef_name,
            'prefix': '',
            'validators': [
                _validate_required_field
            ]
        },
        {
            'field_name': 'job_definition_type',
            'field_value': jobdef_meta.get('job_definition_type'),
            'field_options': JOB_DEFINITION_TYPES,
            'prefix': '',
            'required': True,
            'validators': [
                _validate_options_field
            ]
        },
        {
            'field_name': 'parameters',
            'field_value': jobdef_meta.get('parameters'),
            'prefix': '',
            'required_type': dict,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'retry_strategy',
            'field_value': jobdef_meta.get('retry_strategy'),
            'prefix': '',
            'required_type': dict,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'propagate_tags',
            'field_value': jobdef_meta.get('propagate_tags'),
            'prefix': '',
            'required_type': bool,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'timeout',
            'field_value': jobdef_meta.get('timeout'),
            'prefix': '',
            'required_type': dict,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'tags',
            'field_value': jobdef_meta.get('tags'),
            'prefix': '',
            'required_type': dict,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'platform_capabilities',
            'field_value': jobdef_meta.get('platform_capabilities'),
            'prefix': '',
            'required_type': list,
            'validators': [
                _validate_field_type
            ]
        }
    ]

    job_definition_type = jobdef_meta.get('job_definition_type')
    container_properties = jobdef_meta.get('container_properties')
    node_properties = jobdef_meta.get('node_properties')

    _process_config(jobdef_config)

    if job_definition_type == 'container':
        if not container_properties and not node_properties:
            raise AssertionError(
                "Either 'container_properties' or 'node_properties' must be specified "
                "for 'container' job definition type."
            )
    if container_properties:
        _validate_container_properties(container_properties)
    if node_properties:
        _validate_node_properties(node_properties)


def _validate_container_properties(container_properties, prefix=None):
    """
    Performs check of Batch Job Definition container properties.

    :param container_properties: container properties definition
    :param prefix: prefix that will be displayed before the variable
     name in case of invalidity.

    :raises AssertionError in case of invalidity.

    :return: None
    """
    if not prefix:
        prefix = 'container_properties'

    container_config = [
        {
            'field_name': 'image',
            'field_value': container_properties.get('image'),
            'prefix': prefix,
            'required_type': str,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'vcpus',
            'field_value': container_properties.get('vcpus'),
            'prefix': prefix,
            'required_type': int,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'memory',
            'field_value': container_properties.get('memory'),
            'prefix': prefix,
            'required_type': int,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'command',
            'field_value': container_properties.get('command'),
            'prefix': prefix,
            'required_type': list,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'job_role_arn',
            'field_value': container_properties.get('job_role_arn'),
            'prefix': prefix,
            'required_type': str,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'execution_role_arn',
            'field_value': container_properties.get('execution_role_arn'),
            'prefix': prefix,
            'required_type': str,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'volumes',
            'field_value': container_properties.get('volumes'),
            'prefix': prefix,
            'required_type': list,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'volumes',
            'field_value': container_properties.get('volumes'),
            'prefix': prefix,
            'required_type': list,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'environment',
            'field_value': container_properties.get('environment'),
            'prefix': prefix,
            'required_type': list,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'mount_points',
            'field_value': container_properties.get('mount_points'),
            'prefix': prefix,
            'required_type': list,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'readonly_root_filesystem',
            'field_value': container_properties.get('readonly_root_filesystem'),
            'prefix': prefix,
            'required_type': bool,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'privileged',
            'field_value': container_properties.get('privileged'),
            'prefix': prefix,
            'required_type': bool,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'ulimits',
            'field_value': container_properties.get('ulimits'),
            'prefix': prefix,
            'required_type': list,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'user',
            'field_value': container_properties.get('user'),
            'prefix': prefix,
            'required_type': str,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'instance_type',
            'field_value': container_properties.get('instance_type'),
            'prefix': prefix,
            'required_type': str,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'resource_requirements',
            'field_value': container_properties.get('resource_requirements'),
            'prefix': prefix,
            'required_type': list,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'linux_parameters',
            'field_value': container_properties.get('linux_parameters'),
            'prefix': prefix,
            'required_type': dict,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'log_configuration',
            'field_value': container_properties.get('log_configuration'),
            'prefix': prefix,
            'required_type': dict,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'secrets',
            'field_value': container_properties.get('secrets'),
            'prefix': prefix,
            'required_type': list,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'network_configuration',
            'field_value': container_properties.get('network_configuration'),
            'prefix': prefix,
            'required_type': dict,
            'validators': [
                _validate_field_type
            ]
        },
        {
            'field_name': 'fargate_platform_configuration',
            'field_value': container_properties.get('fargate_platform_configuration'),
            'prefix': prefix,
            'required_type': dict,
            'validators': [
                _validate_field_type
            ]
        },
    ]
    _process_config(container_config)


def _validate_node_properties(node_properties):
    """
    Performs check of Batch Job Definition node properties.
    :param node_properties: container properties definition

    :raises AssertionError in case of invalidity.

    :return: None
    """
    prefix = "node_properties"

    node_config = [
        {
            'field_name': 'num_nodes',
            'field_value': node_properties.get('num_nodes'),
            'prefix': prefix,
            'required_type': int,
            'validators': [
                _validate_field_type,
                _validate_required_field
            ]
        },
        {
            'field_name': 'main_node',
            'field_value': node_properties.get('main_node'),
            'prefix': prefix,
            'required_type': int,
            'validators': [
                _validate_field_type,
                _validate_required_field
            ]
        },
        {
            'field_name': 'node_range_properties',
            'field_value': node_properties.get('node_range_properties'),
            'prefix': prefix,
            'required_type': list,
            'validators': [
                _validate_field_type,
                _validate_required_field
            ]
        },
    ]
    _process_config(node_config)

    node_range_properties = node_properties.get('node_range_properties')

    node_range_prefix = prefix + "__node_range_properties"
    for node in node_range_properties:
        container_properties = node.get('container')
        _validate_required_field(
            field_name='container',
            field_value=container_properties,
            prefix=node_range_prefix
        )
        _validate_field_type(
            field_name='container',
            field_value=node.get('container'),
            prefix=node_range_prefix,
            required_type=dict,
        )

        container_prefix = node_range_prefix + '__container'
        _validate_container_properties(container_properties, prefix=container_prefix)


def _validate_options_field(field_name, field_value, field_options, prefix='compute_resources', required=True,
                            **kwargs):
    """
    Checks whether a field contains value from options list.

    :param field_name: name of field that will be displayed in case of invalidity
    :param field_value: value of field to check
    :param field_options: list of options that field can accept
    :param prefix: prefix that will be displayed before field_name in case of invalidity
    :param required: if field is required and can be empty

    :raises AssertionError in case of invalidity.

    :return: None
    """
    if prefix:
        field_name = prefix + '__' + field_name

    if not required and not field_value:
        return
    if required and not field_value:
        raise AssertionError(
            "Missing required Job Definition field: '{0}'".format(field_name)
        )
    if field_value not in field_options:
        raise AssertionError(
            "Job Definition field: '{0}':'{1}' must be one of the following: {2}"
                .format(field_name, str(field_value), field_options)
        )


def _validate_required_field(field_name, field_value, prefix='', **kwargs):
    """
    Checks whether a field is not empty.

    :param field_name: name of field that will be displayed in case of invalidity
    :param field_value: value of field to check
    :param prefix: prefix that will be displayed before field_name in case of invalidity

    :raises AssertionError in case of invalidity.

    :return: None
    """
    if prefix:
        field_name = prefix + '__' + field_name

    if not field_value:
        raise AssertionError(
            "Missing required Job Definition field: {0}".format(field_name)
        )


def _process_config(config):
    for parameter in config:
        for validator in parameter.get('validators'):
            validator(**parameter)



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from abc import abstractmethod, ABC
from functools import cmp_to_key

from syndicate.commons.log_helper import get_user_logger
from syndicate.core.build.deployment_processor import compare_deploy_resources
from syndicate.core.build.meta_processor import resolve_meta, populate_s3_paths
from syndicate.core.constants import \
    (IAM_POLICY, IAM_ROLE, LAMBDA_TYPE, DYNAMO_TABLE_TYPE, S3_BUCKET_TYPE,
     CLOUD_WATCH_RULE_TYPE, SQS_QUEUE_TYPE, API_GATEWAY_TYPE, SNS_TOPIC_TYPE,
     CLOUD_WATCH_ALARM_TYPE, KINESIS_STREAM_TYPE, COGNITO_FEDERATED_POOL_TYPE,
     SNS_PLATFORM_APPLICATION_TYPE, BATCH_COMPENV_TYPE, BATCH_JOBQUEUE_TYPE,
     BATCH_JOBDEF_TYPE, LAMBDA_LAYER_TYPE, EVENT_BRIDGE_RULE_TYPE)

_LOG = get_user_logger()


class BuildMetaTransformer(ABC):

    def __init__(self, bundle_name=None):
        from syndicate.core import CONFIG, RESOURCES_PROVIDER
        self.config = CONFIG
        self.resources_provider = RESOURCES_PROVIDER

        self.transformer_mapping = {
            IAM_POLICY: self._transform_iam_managed_policy,
            IAM_ROLE: self._transform_iam_role,
            LAMBDA_TYPE: self._transform_lambda,
            LAMBDA_LAYER_TYPE: self._transform_lambda_layer,
            DYNAMO_TABLE_TYPE: self._transform_dynamo_db_table,
            S3_BUCKET_TYPE: self._transform_s3_bucket,
            CLOUD_WATCH_RULE_TYPE: self._transform_cloud_watch_rule,
            EVENT_BRIDGE_RULE_TYPE: self._transform_cloud_watch_rule,
            SQS_QUEUE_TYPE: self._transform_sqs_queue,
            API_GATEWAY_TYPE: self._transform_api_gateway,
            SNS_TOPIC_TYPE: self._transform_sns_topic,
            SNS_PLATFORM_APPLICATION_TYPE: self._transform_sns_application,
            CLOUD_WATCH_ALARM_TYPE: self._transform_cloudwatch_alarm,
            KINESIS_STREAM_TYPE: self._transform_kinesis_stream,
            COGNITO_FEDERATED_POOL_TYPE: self._transform_cognito,
            BATCH_COMPENV_TYPE: self._transform_batch_compenv,
            BATCH_JOBQUEUE_TYPE: self._transform_batch_jobqueue,
            BATCH_JOBDEF_TYPE: self._transform_batch_jobdef
        }
        self.bundle_name = bundle_name

    def transform_build_meta(self, build_meta):
        build_meta = resolve_meta(build_meta)
        build_meta = populate_s3_paths(build_meta, self.bundle_name)
        resources_list = list(build_meta.items())
        resources_list.sort(key=cmp_to_key(compare_deploy_resources))
        for name, resource in resources_list:
            resource_type = resource.get('resource_type')
            transformer = self.transformer_mapping.get(resource_type)
            if transformer is None:
                _LOG.warning("Transformation is not supported for resources "
                    "of the '{}' type".format(resource_type))
                continue
            transformer(name=name, resource=resource)
        return self._compose_template()

    @abstractmethod
    def output_file_name(self) -> str:
        pass

    @abstractmethod
    def _transform_iam_managed_policy(self, name, resource):
        pass

    @abstractmethod
    def _transform_iam_role(self, name, resource):
        pass

    @abstractmethod
    def _transform_lambda(self, name, resource):
        pass

    @abstractmethod
    def _transform_lambda_layer(self, name, resource):
        pass

    @abstractmethod
    def _transform_dynamo_db_table(self, name, resource):
        pass

    @abstractmethod
    def _transform_s3_bucket(self, name, resource):
        pass

    @abstractmethod
    def _transform_cloud_watch_rule(self, name, resource):
        pass

    @abstractmethod
    def _transform_api_gateway(self, name, resource):
        pass

    @abstractmethod
    def _transform_sns_topic(self, name, resource):
        pass

    @abstractmethod
    def _transform_cloudwatch_alarm(self, name, resource):
        pass

    @abstractmethod
    def _transform_sqs_queue(self, name, resource):
        pass

    @abstractmethod
    def _transform_kinesis_stream(self, name, resource):
        pass

    @abstractmethod
    def _transform_sns_application(self, name, resource):
        pass

    @abstractmethod
    def _transform_cognito(self, name, resource):
        pass

    @abstractmethod
    def _transform_batch_compenv(self, name, resource):
        pass

    @abstractmethod
    def _transform_batch_jobqueue(self, name, resource):
        pass

    @abstractmethod
    def _transform_batch_jobdef(self, name, resource):
        pass

    @abstractmethod
    def _compose_template(self):
        pass



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.core.resources.batch_jobqueue_resource import DEFAULT_STATE
from syndicate.core.transform.terraform.converter.tf_resource_converter import \
    TerraformResourceConverter
from syndicate.core.transform.terraform.tf_resource_reference_builder import \
    build_com_env_arn


class BatchJobQueueEnvConverter(TerraformResourceConverter):

    def convert(self, name, resource):
        state = resource.get('state')
        if not state:
            state = DEFAULT_STATE
        priority = resource.get('priority')
        compute_order = resource.get('compute_environment_order', [])
        batch_queue_env = job_queue(name=name, state=state, priority=priority,
                                    compute_environment_order=compute_order)
        self.template.add_aws_batch_job_queue(meta=batch_queue_env)


def job_queue(name, state, priority, compute_environment_order):
    params = {
        'name': name,
        'state': state,
        'priority': priority
    }

    com_env_order = []
    compute_environment_order.sort(key=lambda o: o['order'], reverse=False)
    for order_def in compute_environment_order:
        compute_env = order_def['compute_environment']
        com_env_order.append(build_com_env_arn(com_env_name=compute_env))

    if com_env_order:
        params['compute_environments'] = com_env_order

    resource = {
        name: params
    }
    return resource



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import time
from hashlib import md5

from botocore.exceptions import ClientError

from syndicate.commons import deep_get
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import (
    SOURCE_ARN_DEEP_KEY, SECURITY_SCHEMAS_DEEP_KEY,
    API_GW_DEFAULT_THROTTLING_RATE_LIMIT,
    API_GW_DEFAULT_THROTTLING_BURST_LIMIT
)
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import (build_description_obj,
                                             validate_params)
from syndicate.connection.api_gateway_connection import ApiGatewayV2Connection, \
    ApiGatewayConnection
from syndicate.core.resources.lambda_resource import LambdaResource

API_REQUIRED_PARAMS = ['resources', 'deploy_stage']

_LOG = get_logger('syndicate.core.resources.api_gateway_resource')
USER_LOG = get_user_logger()

SUPPORTED_METHODS = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE', 'OPTIONS',
                     'HEAD', 'ANY']
_CORS_HEADER_NAME = 'Access-Control-Allow-Origin'
_CORS_HEADER_VALUE = "'*'"
_COGNITO_AUTHORIZER_TYPE = 'COGNITO_USER_POOLS'
_CUSTOM_AUTHORIZER_TYPE = 'CUSTOM'

X_SDCT_EXTENSION_KEY = 'x-syndicate-cognito-userpool-names'
PROVIDER_ARNS_KEY = 'providerARNs'

POLICY_STATEMENT_SINGLETON = 'policy_statement_singleton'

_REQUEST_VALIDATORS = {
    'NONE': {
        'validate_request_body': False,
        'validate_request_parameters': False,
        'id': None
    },
    'Validate body': {
        'validate_request_body': True,
        'validate_request_parameters': False,
        'id': None
    },
    'Validate query string parameters and headers': {
        'validate_request_parameters': True,
        'validate_request_body': False,
        'id': None
    },
    'Validate body, query string parameters, and headers': {
        'validate_request_body': True,
        'validate_request_parameters': True,
        'id': None
    }
}

_DISABLE_THROTTLING_VALUE = -1


class ApiGatewayResource(BaseResource):

    def __init__(self, apigw_conn: ApiGatewayConnection,
                 apigw_v2_conn: ApiGatewayV2Connection,
                 lambda_res: LambdaResource,
                 cognito_res, account_id, region) -> None:
        self.connection = apigw_conn
        self.lambda_res = lambda_res
        self.cognito_res = cognito_res
        self.account_id = account_id
        self.region = region
        self.apigw_v2 = apigw_v2_conn

    def _create_default_validators(self, api_id):
        for name, options in _REQUEST_VALIDATORS.items():
            _id = self.connection.create_request_validator(
                api_id, name, options['validate_request_body'],
                options['validate_request_parameters']
            )
            options['id'] = _id

    def _retrieve_request_validator_id(self, api_id, request_validator=None):
        request_validator = request_validator or {}
        if not request_validator:
            return
        validate_request_body = request_validator.get(
            'validate_request_body') or False
        validate_request_parameters = request_validator.get(
            'validate_request_parameters') or False
        name = request_validator.get('name')

        if name:
            return self.connection.create_request_validator(
                api_id, name, validate_request_body,
                validate_request_parameters)
        for validators in _REQUEST_VALIDATORS.values():
            if (validators['validate_request_body'] == validate_request_body) \
                    and (validators[
                             'validate_request_parameters'] == validate_request_parameters):
                return validators['id']

    def api_resource_identifier(self, name, output=None):
        if output:
            # api currently is not located in different regions
            # process only first object
            api_output = list(output.items())[0][1]
            # find id from the output
            return api_output['description']['id']
        # if output is not provided - try to get API by name
        # cause there is no another option
        return self.connection.get_api_id(name)

    def create_api_gateway(self, args):
        """ Create api gateway in pool in sub processes.
    
        :type args: list
        """
        return self.create_pool(self._create_api_gateway_from_meta, args, 1)

    def create_api_gateway_openapi(self, args):
        """ Create OpenAPI api gateway in pool in sub processes.

        :type args: list
        """
        return self.create_pool(self._create_api_gateway_openapi_from_meta,
                                args, 1)

    def api_gateway_update_processor(self, args):
        return self.create_pool(self._create_or_update_api_gateway, args, 1)

    def update_api_gateway_openapi(self, args):
        return self.create_pool(self._update_api_gateway_openapi_from_meta, args, 1)

    @unpack_kwargs
    def _create_or_update_api_gateway(self, name, meta,
                                      current_configurations):
        if current_configurations:
            # api currently is not located in different regions
            # process only first object
            api_output = list(current_configurations.items())[0][1]
            # find id from the output
            api_id = api_output['description']['id']
            # check that api does not exist
            api_response = self.connection.get_api(api_id)
            if api_response:
                # find all existing resources
                existing_resources = api_output['description']['resources']
                existing_paths = [i['path'] for i in existing_resources]
                meta_api_resources = meta['resources']
                resources_statement_singleton = meta.get(
                    POLICY_STATEMENT_SINGLETON)
                api_resp = meta.get('api_method_responses')
                api_integration_resp = meta.get(
                    'api_method_integration_responses')
                api_resources = {}
                for resource_path, resource_meta in meta_api_resources.items():
                    if resource_path not in existing_paths:
                        api_resources[resource_path] = resource_meta
                if api_resources:
                    _LOG.debug(
                        'Going to continue deploy API Gateway {0} ...'.format(
                            api_id))
                    args = self.__prepare_api_resources_args(
                        api_id=api_id,
                        api_resources=api_resources,
                        api_resp=api_resp,
                        api_integration_resp=api_integration_resp,
                        resources_statement_singleton=resources_statement_singleton)
                    self.create_pool(self._create_resource_from_metadata,
                                     args, 1)
                    # add headers
                    # waiter b4 customization
                    time.sleep(10)
                    _LOG.debug(
                        'Customizing API Gateway {0} responses...'.format(
                            api_id))
                else:
                    # all resources created, but need to override
                    api_resources = meta_api_resources
                # _customize_gateway_responses call is commented due to
                # botocore InternalFailure while performing the call.
                # will be fixed later
                # _customize_gateway_responses(api_id)
                # deploy api
                _LOG.debug('Deploying API Gateway {0} ...'.format(api_id))
                self.__deploy_api_gateway(api_id, meta, api_resources)
                return self.describe_api_resources(api_id=api_id, meta=meta,
                                                   name=name)
            else:
                # api does not exist, so create a new
                return self._create_api_gateway_from_meta(
                    {'name': name, 'meta': meta})
        else:
            # object is not present, so just create a new api
            return self._create_api_gateway_from_meta(
                {'name': name, 'meta': meta})

    def _escape_path(self, parameter):
        index = parameter.find('/', 0)
        if index == -1:
            return parameter
        parameter = parameter[:index] + '~1' + parameter[index + 1:]
        return self._escape_path(parameter)

    def configure_resources(self, api_id, stage_name, api_resources):
        for resource_path, resource_meta in api_resources.items():
            for method_name, method_meta in resource_meta.items():
                if method_name in SUPPORTED_METHODS:
                    cache_configuration = method_meta.get(
                        'cache_configuration')
                    throttling_configuration = method_meta.get(
                        'throttling_configuration')
                    cache_ttl_setting = cache_configuration.get(
                        'cache_ttl_sec') if cache_configuration else None
                    encrypt_cache_data = cache_configuration.get(
                        'encrypt_cache_data') if cache_configuration else None
                    throttling_enabled = throttling_configuration.get(
                        'throttling_enabled') if throttling_configuration \
                        else None
                    throttling_rate_limit = throttling_configuration.get(
                        'throttling_rate_limit',
                        API_GW_DEFAULT_THROTTLING_RATE_LIMIT) if (
                        throttling_configuration and throttling_enabled) else \
                        _DISABLE_THROTTLING_VALUE
                    throttling_burst_limit = throttling_configuration.get(
                        'throttling_burst_limit',
                        API_GW_DEFAULT_THROTTLING_BURST_LIMIT) if (
                        throttling_configuration and throttling_enabled) else \
                        _DISABLE_THROTTLING_VALUE
                    patch_operations = []
                    escaped_resource = self._escape_path(resource_path)
                    if cache_ttl_setting is not None:
                        _LOG.info(
                            'Configuring cache for {0}; TTL: {1}'.format(
                                resource_path, cache_ttl_setting))
                        patch_operations = [
                            {
                                'op': 'replace',
                                'path': '/{0}/{1}/caching/ttlInSeconds'.format(
                                    escaped_resource,
                                    method_name),
                                'value': str(cache_ttl_setting),
                            },
                            {
                                'op': 'replace',
                                'path': '/{0}/{1}/caching/enabled'.format(
                                    escaped_resource,
                                    method_name),
                                'value': 'True',
                            }
                        ]
                        if encrypt_cache_data is not None:
                            patch_operations.append({
                                'op': 'replace',
                                'path': '/{0}/{1}/caching/dataEncrypted'.format(
                                    escaped_resource,
                                    method_name),
                                'value': 'true' if bool(
                                    encrypt_cache_data) else 'false'
                            })
                    if throttling_enabled:
                        _LOG.info(
                            'Configuring throttling for {0}; rateLimit: {1}; '
                            'burstLimit: {2}'.format(
                                resource_path, throttling_rate_limit,
                                throttling_burst_limit))
                    else:
                        _LOG.info('Throttling for {0} disabled.'.format(
                            resource_path))
                    patch_operations.append({
                            'op': 'replace',
                            'path': '/{0}/{1}/throttling/rateLimit'
                                    ''.format(escaped_resource,
                                              method_name),
                            'value': str(throttling_rate_limit),
                    })
                    patch_operations.append({
                            'op': 'replace',
                            'path': '/{0}/{1}/throttling/burstLimit'
                                    ''.format(escaped_resource,
                                              method_name),
                            'value': str(throttling_burst_limit),
                    })
                    self.connection.update_configuration(
                        rest_api_id=api_id,
                        stage_name=stage_name,
                        patch_operations=patch_operations
                    )
                    _LOG.info(
                        'Resource {0} was configured'.format(
                            resource_path))

    @unpack_kwargs
    def _create_api_gateway_from_meta(self, name, meta):
        """ Create API Gateway with all specified meta.
    
        :type name: str
        :type meta: dict
        """
        validate_params(name, meta, API_REQUIRED_PARAMS)

        api_resources = meta['resources']
        # whether to put a wildcard in lambda resource-based policy permissions
        resources_permission_singleton = meta.get(POLICY_STATEMENT_SINGLETON)
        # api_gw_describe = self.describe_api_resources(name, meta)
        # if api_gw_describe:
        #     _LOG.info(f'Api gateway with name \'{name}\' exists. Returning')
        #     return api_gw_describe
        # _LOG.info(f'Api gateway with name \'{name}\' does not exist. Creating')
        api_item = self.connection.create_rest_api(
            api_name=name,
            binary_media_types=meta.get('binary_media_types'))
        api_id = api_item['id']

        # create default request validators
        self._create_default_validators(api_id)

        # set minimumCompressionSize if the param exists
        minimum_compression_size = meta.get('minimum_compression_size', None)
        if not minimum_compression_size:
            _LOG.debug("No minimal_compression_size param - "
                       "compression isn't enabled")
        self.connection.update_compression_size(
            rest_api_id=api_id,
            compression_size=minimum_compression_size)

        # deploy authorizers
        authorizers = meta.get('authorizers', {})
        for key, val in authorizers.items():
            uri = None
            provider_arns = []
            if val.get('type') == _COGNITO_AUTHORIZER_TYPE:
                for pool in val.get('user_pools'):
                    user_pool_id = self.cognito_res.get_user_pool_id(pool)
                    provider_arns.append(
                        f'arn:aws:cognito-idp:{self.region}:{self.account_id}:'
                        f'userpool/{user_pool_id}')
            else:
                lambda_version = val.get('lambda_version')
                lambda_name = val.get('lambda_name')
                lambda_alias = val.get('lambda_alias')
                lambda_arn = self.lambda_res. \
                    resolve_lambda_arn_by_version_and_alias(lambda_name,
                                                            lambda_version,
                                                            lambda_alias)
                uri = f'arn:aws:apigateway:{self.region}:lambda:path/' \
                      f'2015-03-31/functions/{lambda_arn}/invocations'
                api_source_arn = (f'arn:aws:execute-api:{self.region}:'
                                  f'{self.account_id}:{api_id}/*/*')
                self.lambda_res.add_invocation_permission(
                    statement_id=api_id,
                    name=lambda_arn,
                    source_arn=api_source_arn,
                    principal='apigateway.amazonaws.com')

            self.connection.create_authorizer(api_id=api_id, name=key,
                                              type=val['type'],
                                              authorizer_uri=uri,
                                              identity_source=val.get(
                                                  'identity_source'),
                                              ttl=val.get('ttl'),
                                              provider_arns=provider_arns)

        models = meta.get('models')
        if models:
            args = [{'api_id': api_id, 'models': {k: v}} for k, v in
                    models.items()]
            self.create_pool(self._create_model_from_metadata, args, 1)
        if api_resources:
            api_resp = meta.get('api_method_responses')
            api_integration_resp = meta.get('api_method_integration_responses')
            args = self.__prepare_api_resources_args(
                api_id, api_resources, api_resp, api_integration_resp,
                resources_permission_singleton)
            self.create_pool(self._create_resource_from_metadata, args, 1)
        else:
            _LOG.info('There is no resources in %s API Gateway description.',
                      name)
        # add headers
        # waiter b4 customization
        time.sleep(10)
        _LOG.debug('Customizing API Gateway responses...')
        # _customize_gateway_responses call is commented due to botocore
        # InternalFailure while performing the call. will be fixed later
        # _customize_gateway_responses(api_id)
        # deploy api
        self.__deploy_api_gateway(api_id, meta, api_resources)
        return self.describe_api_resources(api_id=api_id, meta=meta, name=name)

    @unpack_kwargs
    def _create_api_gateway_openapi_from_meta(self, name, meta):
        openapi_context = meta.get('definition')
        deploy_stage = meta.get('deploy_stage')

        self._resolve_cup_ids(openapi_context)

        api_id = self.connection.create_openapi(openapi_context)
        self.connection.deploy_api(api_id, deploy_stage)

        api_lambdas_arns = self.extract_api_gateway_lambdas_arns(
            openapi_context)
        self.create_lambdas_permissions(api_id, api_lambdas_arns, '/*/*/*')
        api_lambda_auth_arns = self.extract_api_gateway_lambda_auth_arns(
            openapi_context)
        self.create_lambdas_permissions(api_id, api_lambda_auth_arns, '/*/*')

        return self.describe_api_resources(api_id=api_id, meta=meta, name=name)

    @unpack_kwargs
    def _update_api_gateway_openapi_from_meta(self, name, meta, context):
        api_id = self.connection.get_api_id(name)
        openapi_context = meta.get('definition')
        deploy_stage = meta.get('deploy_stage')

        self._resolve_cup_ids(openapi_context)

        self.connection.update_openapi(api_id, openapi_context)
        self.connection.deploy_api(api_id, deploy_stage)

        api_lambdas_arns = self.extract_api_gateway_lambdas_arns(
            openapi_context
        )
        self.create_lambdas_permissions(api_id, api_lambdas_arns, '/*/*/*')
        api_lambda_auth_arns = self.extract_api_gateway_lambda_auth_arns(
            openapi_context
        )
        self.create_lambdas_permissions(api_id, api_lambda_auth_arns, '/*/*')

        return self.describe_api_resources(api_id=api_id, meta=meta, name=name)

    def _resolve_cup_ids(self, openapi_context):
        _LOG.debug('Going to resolve Cognito User Pools ARNs')
        security_schemes = \
            openapi_context.get('components', {}).get('securitySchemes', {})

        authorizers = [
            value['x-amazon-apigateway-authorizer']
            for _, value in security_schemes.items()
            if (value.get('x-amazon-apigateway-authtype') ==
                _COGNITO_AUTHORIZER_TYPE.lower()
                and 'x-amazon-apigateway-authorizer' in value)]

        for authorizer in authorizers:
            pools_names = provider_arns = None
            if authorizer.get('type') == _COGNITO_AUTHORIZER_TYPE.lower():
                pools_names = authorizer.get(X_SDCT_EXTENSION_KEY)
                provider_arns = authorizer.get(PROVIDER_ARNS_KEY)
            new_provider_arns = []
            if pools_names:
                for pool_name in pools_names:
                    _LOG.debug(f'Resolving ARN for Cognito User Pool by name '
                               f'{pool_name}')
                    pool_id = self.cognito_res.get_user_pool_id(pool_name)
                    if pool_id:
                        new_provider_arns.append(
                            f'arn:aws:cognito-idp:{self.region}:'
                            f'{self.account_id}:userpool/{pool_id}')
                    else:
                        USER_LOG.warn(f'Can\'t resolve Cognito User Pool ID '
                                      f'by name "{pool_name}"! For more '
                                      f'details see syndicate logs.')
            elif provider_arns:
                for arn in provider_arns:
                    pool_id = arn.split('/')[-1]
                    _LOG.debug(f'Resolving ARN for Cognito User Pool by ID '
                               f'{pool_id}')
                    if self.cognito_res.is_user_pool_exists(pool_id):
                        new_provider_arns.append(arn)
                    else:
                        USER_LOG.warn(f'Cognito User Pool with ID {pool_id} '
                                      f'not found.')

            if new_provider_arns:
                _LOG.debug(f'Going to apply the next provider ARNs '
                           f'{new_provider_arns} to API Gateway '
                           f'{deep_get(openapi_context, ["info", "title"])}')
                authorizer[PROVIDER_ARNS_KEY] = new_provider_arns
            else:
                raise AssertionError(
                    f'Cognito User Pools can\'t be resolved by ' + 'names: '
                    f'{pools_names}' if pools_names else
                    f'ARNs: {provider_arns}')

    def get_lambda_permissions_for_api(self, lambda_arn, api_gateway_id):
        permissions = self.lambda_res.get_existing_permissions(lambda_arn)
        # Filter the permissions related to the specific API Gateway
        filtered_permissions = [
            statement for statement in permissions
            if deep_get(
                statement, SOURCE_ARN_DEEP_KEY, ""
            ).startswith('arn:aws:execute-api:') and api_gateway_id in deep_get(
                statement, SOURCE_ARN_DEEP_KEY, "")
        ]

        return filtered_permissions

    def create_lambdas_permissions(
        self,
        api_gateway_id: str,
        api_lambdas_arns: set[str],
        route: str
    ):

        api_source_arn = (f'arn:aws:execute-api:{self.region}:'
                          f'{self.account_id}:{api_gateway_id}{route}')
        for lambda_arn in api_lambdas_arns:
            _id = f'{lambda_arn}-{api_source_arn}'
            statement_id = md5(_id.encode('utf-8')).hexdigest()
            self.lambda_res.add_invocation_permission(
                name=lambda_arn,
                principal='apigateway.amazonaws.com',
                source_arn=api_source_arn,
                statement_id=statement_id,
                exists_ok=True
            )

    def remove_lambdas_permissions(self, api_gateway_id, api_lambdas_arns):
        for lambda_arn in api_lambdas_arns:
            existing_permissions = self.get_lambda_permissions_for_api(
                lambda_arn, api_gateway_id)

            existing_permissions = {
                deep_get(perm, SOURCE_ARN_DEEP_KEY): perm.get('Sid')
                for perm in existing_permissions
            }
            self.lambda_res.remove_permissions(lambda_arn,
                                               existing_permissions.values())

    @staticmethod
    def get_deploy_stage_name(stage_name=None):
        return stage_name if stage_name else 'prod'

    def __deploy_api_gateway(self, api_id, meta, api_resources):
        deploy_stage = self.get_deploy_stage_name(meta.get('deploy_stage'))
        cache_cluster_configuration = meta.get('cluster_cache_configuration')
        root_cache_enabled = cache_cluster_configuration.get(
            'cache_enabled') if cache_cluster_configuration else None
        cache_size = cache_cluster_configuration.get(
            'cache_size') if cache_cluster_configuration else None
        self.connection.deploy_api(api_id, stage_name=deploy_stage,
                                   cache_cluster_enabled=root_cache_enabled,
                                   cache_cluster_size=str(
                                       cache_size) if cache_size else None)
        throttling_cluster_configuration = meta.get(
            'cluster_throttling_configuration')
        throttling_enabled = throttling_cluster_configuration.get(
            'throttling_enabled') if throttling_cluster_configuration else None
        patch_operations = []
        if not throttling_enabled:
            patch_operations.append({
                'op': 'replace',
                'path': '/*/*/throttling/rateLimit',
                'value': str(_DISABLE_THROTTLING_VALUE),
            })
            patch_operations.append({
                'op': 'replace',
                'path': '/*/*/throttling/burstLimit',
                'value': str(_DISABLE_THROTTLING_VALUE),
            })
        # configure caching
        if root_cache_enabled:
            _LOG.debug('Cluster cache configuration found:{0}'.format(
                cache_cluster_configuration))
            # set default ttl for root endpoint
            cluster_cache_ttl_sec = cache_cluster_configuration.get(
                'cache_ttl_sec')
            encrypt_cache_data = cache_cluster_configuration.get(
                'encrypt_cache_data')
            if cluster_cache_ttl_sec is not None:
                patch_operations.append({
                    'op': 'replace',
                    'path': '/*/*/caching/ttlInSeconds',
                    'value': str(cluster_cache_ttl_sec),
                })
            if encrypt_cache_data is not None:
                patch_operations.append({
                    'op': 'replace',
                    'path': '/*/*/caching/dataEncrypted',
                    'value': 'true' if bool(encrypt_cache_data) else 'false'
                })
        # configure throttling
        if throttling_enabled:
            throttling_rate_limit = throttling_cluster_configuration.get(
                'throttling_rate_limit', API_GW_DEFAULT_THROTTLING_RATE_LIMIT)
            throttling_burst_limit = throttling_cluster_configuration.get(
                'throttling_burst_limit', API_GW_DEFAULT_THROTTLING_BURST_LIMIT)
            patch_operations.append({
                'op': 'replace',
                'path': '/*/*/throttling/rateLimit',
                'value': str(throttling_rate_limit),
            })
            patch_operations.append({
                'op': 'replace',
                'path': '/*/*/throttling/burstLimit',
                'value': str(throttling_burst_limit),
            })
        if any([root_cache_enabled, throttling_enabled]):
            self.connection.update_configuration(
                    rest_api_id=api_id,
                    stage_name=deploy_stage,
                    patch_operations=patch_operations
                )
        # customize settings for endpoints
        self.configure_resources(api_id, deploy_stage, api_resources)

    def __prepare_api_resources_args(
            self, api_id, api_resources, api_resp=None,
            api_integration_resp=None,
            resources_statement_singleton: bool = False):
        # describe authorizers and create a mapping
        authorizers = self.connection.get_authorizers(api_id)
        authorizers_mapping = {x['name']: x['id'] for x in authorizers}
        args = []
        for each in api_resources:
            resource_meta = api_resources[each]
            _LOG.info('Creating resource %s ...', each)
            if each.startswith('/'):
                resource_id = self.connection.get_resource_id(api_id, each)
                if resource_id:
                    _LOG.info('Resource %s exists.', each)
                    enable_cors = resource_meta.get('enable_cors')
                    self._check_existing_methods(
                        api_id=api_id, resource_id=resource_id,
                        resource_path=each,
                        resource_meta=resource_meta,
                        enable_cors=enable_cors,
                        authorizers_mapping=authorizers_mapping,
                        api_resp=api_resp,
                        api_integration_resp=api_integration_resp,
                        resources_statement_singleton=resources_statement_singleton)
                else:
                    args.append({
                        'api_id': api_id,
                        'resource_path': each,
                        'resource_meta': resource_meta,
                        'authorizers_mapping': authorizers_mapping,
                        'resources_statement_singleton': resources_statement_singleton
                    })
            else:
                raise AssertionError(
                    "API resource must starts with '/', but found %s", each)
        return args

    def describe_api_resources(self, name, meta, api_id=None):
        if not api_id:
            api = self.connection.get_api_by_name(name)
            if not api:
                return
            api_id = api['id']

        response = self.connection.get_api(api_id)
        if not response:
            return
        response['resources'] = self.connection.get_resources(api_id)
        _LOG.info('Described %s API Gateway.', name)
        arn = 'arn:aws:apigateway:{0}::/restapis/{1}'.format(self.region,
                                                             api_id)
        return {
            arn: build_description_obj(response, name, meta)
        }

    def describe_openapi(self, api_id, stage_name):
        response = self.connection.describe_openapi(api_id, stage_name)
        return json.loads(response['body'].read().decode("utf-8")) \
            if isinstance(response, dict) else None

    def _check_existing_methods(self, api_id, resource_id, resource_path,
                                resource_meta,
                                enable_cors, authorizers_mapping,
                                api_resp=None,
                                api_integration_resp=None,
                                resources_statement_singleton: bool = False
                                ):
        """ Check if all specified methods exist and create some if not.
    
        :type api_id: str
        :type resource_id: str
        :type resource_meta: dict
        :type enable_cors: bool or None
        :type:
        """
        methods_statement_singleton = resource_meta.get(
            POLICY_STATEMENT_SINGLETON)
        for method in resource_meta:
            if method == 'enable_cors':
                continue
            if self.connection.get_method(api_id, resource_id, method):
                _LOG.info('Method %s exists.', method)
                continue
            else:
                _LOG.info('Creating method %s for resource %s...',
                          method, resource_id)
                self._create_method_from_metadata(
                    api_id=api_id,
                    resource_id=resource_id,
                    resource_path=resource_path,
                    method=method,
                    method_meta=resource_meta[method],
                    authorizers_mapping=authorizers_mapping,
                    api_resp=api_resp,
                    api_integration_resp=api_integration_resp,
                    enable_cors=enable_cors,
                    resources_statement_singleton=resources_statement_singleton,
                    methods_statement_singleton=methods_statement_singleton
                )
            if enable_cors and not self.connection.get_method(api_id,
                                                              resource_id,
                                                              'OPTIONS'):
                _LOG.info('Enabling CORS for resource %s...', resource_id)
                self.connection.enable_cors_for_resource(api_id, resource_id)

    @unpack_kwargs
    def _create_resource_from_metadata(self, api_id, resource_path,
                                       resource_meta,
                                       authorizers_mapping,
                                       resources_statement_singleton: bool = False):
        self.connection.create_resource(api_id, resource_path)
        _LOG.info('Resource %s created.', resource_path)
        resource_id = self.connection.get_resource_id(api_id, resource_path)
        enable_cors = resource_meta.get('enable_cors')
        methods_statement_singleton = resource_meta.get(
            POLICY_STATEMENT_SINGLETON)
        for method in resource_meta:
            try:
                if method == 'enable_cors' or method not in SUPPORTED_METHODS:
                    continue

                method_meta = resource_meta[method]
                _LOG.info('Creating method %s for resource %s...',
                          method, resource_path)
                self._create_method_from_metadata(
                    api_id=api_id,
                    resource_id=resource_id,
                    resource_path=resource_path,
                    method=method,
                    method_meta=method_meta,
                    enable_cors=enable_cors,
                    authorizers_mapping=authorizers_mapping,
                    resources_statement_singleton=resources_statement_singleton,
                    methods_statement_singleton=methods_statement_singleton
                )
            except Exception as e:
                _LOG.error('Resource: {0}, method {1}.'
                           .format(resource_path, method), exc_info=True)
                raise e
            _LOG.info('Method %s for resource %s created.', method,
                      resource_path)
        # create enable cors only after all methods in resource created
        if enable_cors:
            self.connection.enable_cors_for_resource(api_id, resource_id)
            _LOG.info('CORS enabled for resource %s', resource_path)

    def _create_method_from_metadata(
            self, api_id, resource_id, resource_path, method, method_meta,
            authorizers_mapping, enable_cors=False, api_resp=None,
            api_integration_resp=None,
            resources_statement_singleton: bool = False,
            methods_statement_singleton: bool = False):
        resources_statement_singleton = resources_statement_singleton or False
        methods_statement_singleton = methods_statement_singleton or False
        # init responses for method
        resp = self.init_method_responses(api_resp, method_meta)

        # init integration responses for method
        integr_resp = self.init_integration_method_responses(
            api_integration_resp, method_meta)

        # resolve authorizer if needed
        authorization_type = method_meta.get('authorization_type')
        if authorization_type not in ['NONE', 'AWS_IAM']:
            # type is authorizer, so add id to meta
            authorizer_id = authorizers_mapping.get(authorization_type)
            if not authorizer_id:
                raise AssertionError(
                    'Authorizer {0} does not exist'.format(authorization_type))
            method_meta['authorizer_id'] = authorizer_id
            authorizer = self.connection.get_authorizer(
                api_id, authorizer_id).get('type')
            if authorizer == _COGNITO_AUTHORIZER_TYPE:
                authorization_type = _COGNITO_AUTHORIZER_TYPE
            else:
                authorization_type = _CUSTOM_AUTHORIZER_TYPE

        method_request_models = method_meta.get('method_request_models')
        if method_request_models:
            (content_type, name), = method_request_models.items()
            model = self.connection.get_model(api_id, name)
            method_request_models = model if not model else method_request_models

        request_validator_id = self._retrieve_request_validator_id(
            api_id, method_meta.get('request_validator'))

        self.connection.create_method(
            api_id, resource_id, method,
            authorization_type=authorization_type,
            authorizer_id=method_meta.get('authorizer_id'),
            api_key_required=method_meta.get('api_key_required'),
            request_parameters=method_meta.get('method_request_parameters'),
            request_models=method_request_models,
            request_validator=request_validator_id)
        # second step: create integration
        integration_type = method_meta.get('integration_type')
        # set up integration - lambda or aws service
        body_template = method_meta.get('integration_request_body_template')
        passthrough_behavior = method_meta.get(
            'integration_passthrough_behavior')
        request_parameters = method_meta.get('integration_request_parameters')
        # TODO split to map - func implementation
        if integration_type:
            if integration_type == 'lambda':
                lambda_name = method_meta['lambda_name']
                # alias has a higher priority than version in arn resolving
                lambda_version = method_meta.get('lambda_version')
                lambda_alias = method_meta.get('lambda_alias')
                lambda_arn = self.lambda_res. \
                    resolve_lambda_arn_by_version_and_alias(lambda_name,
                                                            lambda_version,
                                                            lambda_alias)
                enable_proxy = method_meta.get('enable_proxy')
                cache_configuration = method_meta.get('cache_configuration')
                cache_key_parameters = cache_configuration.get(
                    'cache_key_parameters') if cache_configuration else None
                self.connection.create_lambda_integration(
                    lambda_arn, api_id, resource_id, method, body_template,
                    passthrough_behavior, method_meta.get('lambda_region'),
                    enable_proxy=enable_proxy,
                    cache_key_parameters=cache_key_parameters,
                    request_parameters=request_parameters)
                # add permissions to invoke
                # Allows to apply method or resource singleton of a policy
                # statement, setting wildcard on the respective scope.

                api_source_arn = f"arn:aws:execute-api:{self.region}:" \
                                 f"{self.account_id}:{api_id}/*" \
                                 "/{method}/{path}"

                _method, _path = method, resource_path.lstrip('/')
                if resources_statement_singleton:
                    _path = '*'
                if methods_statement_singleton:
                    _method = '*'

                api_source_arn = api_source_arn.format(
                    method=_method, path=_path
                )
                _id = f'{lambda_arn}-{api_source_arn}'
                statement_id = md5(_id.encode('utf-8')).hexdigest()
                response: dict = self.lambda_res.add_invocation_permission(
                    name=lambda_arn,
                    principal='apigateway.amazonaws.com',
                    source_arn=api_source_arn,
                    statement_id=statement_id,
                    exists_ok=True
                )
                if response is None:
                    message = f'Permission: \'{statement_id}\' attached to ' \
                              f'\'{lambda_arn}\' lambda to allow ' \
                              f'lambda:InvokeFunction for ' \
                              f'apigateway.amazonaws.com principal from ' \
                              f'\'{api_source_arn}\' SourceArn already exists.'
                    _LOG.warning(message + ' Skipping.')

            elif integration_type == 'service':
                uri = method_meta.get('uri')
                role = method_meta.get('role')
                integration_method = method_meta.get('integration_method')
                self.connection.create_service_integration(self.account_id,
                                                           api_id,
                                                           resource_id,
                                                           method,
                                                           integration_method,
                                                           role, uri,
                                                           body_template,
                                                           passthrough_behavior,
                                                           request_parameters)
            elif integration_type == 'mock':
                self.connection.create_mock_integration(api_id, resource_id,
                                                        method,
                                                        body_template,
                                                        passthrough_behavior)
            elif integration_type == 'http':
                integration_method = method_meta.get('integration_method')
                uri = method_meta.get('uri')
                enable_proxy = method_meta.get('enable_proxy')
                self.connection.create_http_integration(api_id, resource_id,
                                                        method,
                                                        integration_method,
                                                        uri,
                                                        body_template,
                                                        passthrough_behavior,
                                                        enable_proxy)
            else:
                raise AssertionError('%s integration type does not exist.',
                                     integration_type)
        # third step: setup method responses
        if resp:
            for response in resp:
                self.connection.create_method_response(
                    api_id, resource_id, method, response.get('status_code'),
                    response.get('response_parameters'),
                    response.get('response_models'), enable_cors)
        else:
            self.connection.create_method_response(
                api_id, resource_id, method, enable_cors=enable_cors)
        # fourth step: setup integration responses
        if integr_resp:
            for each in integr_resp:
                self.connection.create_integration_response(
                    api_id, resource_id, method, each.get('status_code'),
                    each.get('error_regex'),
                    each.get('response_parameters'),
                    each.get('response_templates'), enable_cors)
        else:
            self.connection.create_integration_response(
                api_id, resource_id, method, enable_cors=enable_cors)

    @staticmethod
    def init_method_responses(api_resp, method_meta):
        method_responses = method_meta.get("responses")
        if method_responses:
            resp = method_responses
        elif api_resp:
            resp = api_resp
        else:
            resp = []
        return resp

    @staticmethod
    def init_integration_method_responses(api_integration_resp,
                                          method_meta):
        integration_method_responses = method_meta.get("integration_responses")
        if integration_method_responses:
            integration_resp = integration_method_responses
        elif api_integration_resp:
            integration_resp = api_integration_resp
        else:
            integration_resp = []
        return integration_resp

    def _customize_gateway_responses(self, api_id):
        responses = self.connection.get_gateway_responses(api_id)
        response_types = [r['responseType'] for r in responses]
        for response_type in response_types:
            time.sleep(10)
            self.connection.add_header_to_gateway_response(api_id,
                                                           response_type,
                                                           _CORS_HEADER_NAME,
                                                           _CORS_HEADER_VALUE)

    @staticmethod
    def _get_lambdas_invoked_by_api_gw(resources_meta, retrieve_aliases=False):
        lambdas = set()
        for resource, meta in resources_meta.items():
            for method, description in meta.items():
                if method in SUPPORTED_METHODS:
                    lambda_ = description.get('lambda_name')
                    if lambda_:
                        if retrieve_aliases:
                            lambda_ = (lambda_,
                                       description.get('lambda_alias'))
                        lambdas.add(lambda_)
        return list(lambdas)

    def remove_api_gateways(self, args):
        for arg in args:
            self._remove_api_gateway(**arg)
            # wait for success deletion
            time.sleep(60)

    def _remove_invocation_permissions_from_lambdas(self, config):
        api_id = config['description']['id']
        _LOG.info(fr'Removing invocation permissions for api {api_id}')
        lambdas_aliases = self._get_lambdas_invoked_by_api_gw(
            config['resource_meta'].get('resources', {}),
            retrieve_aliases=True)
        for lambda_, alias in lambdas_aliases:
            _LOG.info(f'Removing invocation permissions for api {api_id} '
                      f'from lambda {lambda_} and alias {alias}')
            statements = self.lambda_res.get_invocation_permission(
                lambda_name=self.lambda_res.build_lambda_arn(lambda_),
                qualifier=alias
            ).get('Statement', [])
            ids_to_remove = [st.get('Sid') for st in statements if
                             api_id in st.get('Condition', {}).get(
                                 'ArnLike', {}).get('AWS:SourceArn', '')]
            self.lambda_res.remove_invocation_permissions(
                lambda_name=lambda_, qualifier=alias,
                ids_to_remove=ids_to_remove
            )

    def _remove_api_gateway(self, arn, config):
        api_id = config['description']['id']
        stage_name = config["resource_meta"]["deploy_stage"]
        openapi_context = self.describe_openapi(api_id, stage_name)
        if openapi_context:
            api_lambdas_arns = self.extract_api_gateway_lambdas_arns(
                openapi_context)
            api_lambda_auth_arns = self.extract_api_gateway_lambda_auth_arns(
                openapi_context)
            self.remove_lambdas_permissions(
                api_id,
                {*api_lambdas_arns, *api_lambda_auth_arns}
            )
        try:
            self.connection.remove_api(api_id)
            _LOG.info(f'API Gateway {api_id} was removed.')
        except ClientError as e:
            if e.response['Error']['Code'] == 'NotFoundException':
                _LOG.warning('API Gateway %s is not found', api_id)
            else:
                raise

    @unpack_kwargs
    def _create_model_from_metadata(self, api_id, models):
        _LOG.info('Going to process API Gateway models')
        for name, model_data in models.items():
            description = model_data.get('description')
            schema = model_data.get('schema')
            if isinstance(schema, dict):
                schema = json.dumps(schema)
            content_type = model_data.get('content_type')
            self.connection.create_model(
                api_id, name, content_type, description, schema)

    def build_web_socket_api_gateway_arn(self, api_id: str) -> str:
        return f'arn:aws:execute-api:{self.apigw_v2.client.meta.region_name}' \
               f':{self.account_id}:{api_id}/*'

    def create_web_socket_api_gateway(self, args):
        return self.create_pool(self._create_web_socket_api_from_meta, args, 1)

    @unpack_kwargs
    def _create_web_socket_api_from_meta(self, name: str, meta: dict):
        stage_name = meta.get('deploy_stage')
        resources = meta.get('resources') or {}
        route_selection_expression = meta.get('route_selection_expression')
        api_id = self.apigw_v2.create_web_socket_api(
            name=name, route_selection_expression=route_selection_expression)
        for route_name, route_meta in resources.items():
            int_type = route_meta.get('integration_type') or 'lambda'
            if int_type != 'lambda':
                _LOG.error(f'Integration type: {int_type} currently '
                           f'not supported. Skipping..')
                continue
            lambda_name = route_meta['lambda_name']
            lambda_version = route_meta.get('lambda_version')
            lambda_alias = route_meta.get('lambda_alias')
            lambda_arn = self.lambda_res.resolve_lambda_arn_by_version_and_alias(
                lambda_name, lambda_version, lambda_alias)

            integration_id = self.apigw_v2.create_lambda_integration(
                api_id=api_id,
                lambda_arn=lambda_arn,
                enable_proxy=route_meta.get('enable_proxy') or False
            )
            self.apigw_v2.put_route_integration(
                api_id=api_id,
                route_name=route_name,
                integration_id=integration_id
            )
            source_arn = f'{self.build_web_socket_api_gateway_arn(api_id)}/{route_name}'

            self.lambda_res.add_invocation_permission(
                name=lambda_arn,
                principal='apigateway.amazonaws.com',
                source_arn=source_arn,
                statement_id=f'{name}-{route_name.strip("$")}-invoke',
                exists_ok=True
            )

        self.apigw_v2.create_stage(api_id=api_id, stage_name=stage_name)
        return self.describe_v2_api_gateway(
            name=name, meta=meta, api_id=api_id
        )

    def describe_v2_api_gateway(self, name, meta, api_id=None):
        if not api_id:
            api = self.apigw_v2.get_api_by_name(name)
            if not api:
                return
            api_id = api['ApiId']

        # response = self.connection.get_api(api_id)
        response = {'ApiId': api_id}
        # maybe the arn is not valid - I didn't manage to find a valid
        # example browsing for 5 minutes, so the hell with it. Currently,
        # nothing depends on it
        arn = self.build_web_socket_api_gateway_arn(api_id)
        return {
            arn: build_description_obj(response, name, meta)
        }

    def remove_v2_api_gateway(self, args):
        for arg in args:
            self._remove_v2_api_gateway(**arg)
            # wait for success deletion
            # time.sleep(60)

    def _remove_v2_api_gateway(self, arn, config):
        api_id = config.get('description', {}).get('ApiId')
        if not api_id:
            _LOG.warning('V2 api id not found in output. Skipping')
            return

        lambda_arns = []
        routes = self.apigw_v2.get_routes(api_id)
        for route in routes['Items']:
            lambda_arns.append(self.apigw_v2.get_integration(
                api_id, route['Target'].replace('integrations/', '')))

        self.remove_lambdas_permissions(
            api_id, {*[arn for arn in lambda_arns if arn is not None]})
        self.apigw_v2.delete_api(api_id)
        return

    @staticmethod
    def extract_api_gateway_lambdas_arns(openapi_spec):
        api_gateway_lambdas_arns = {*()}

        for path, path_item in openapi_spec.get('paths', {}).items():
            for method, method_data in path_item.items():
                integration = method_data.get('x-amazon-apigateway-integration')
                if not integration or not integration.get('uri'):
                    continue
                uri = integration.get('uri')
                try:
                    lambda_arn = uri.split('/functions/')[1].split('/')[0]
                except IndexError:
                    _LOG.warning(f"Invalid lambda arn in integration uri {uri}")
                    continue

                api_gateway_lambdas_arns.add(lambda_arn)
        return api_gateway_lambdas_arns

    @staticmethod
    def extract_api_gateway_lambda_auth_arns(openapi_spec):
        api_gateway_lambdas_arns = {*()}

        security_schemas = deep_get(openapi_spec, SECURITY_SCHEMAS_DEEP_KEY, {})
        for schema_data in security_schemas.values():
            authorizer = schema_data.get("x-amazon-apigateway-authorizer")
            if not authorizer or not authorizer.get("authorizerUri"):
                continue
            uri = authorizer.get("authorizerUri")
            try:
                lambda_arn = uri.split('/functions/')[1].split('/')[0]
            except IndexError:
                _LOG.warning(f"Invalid lambda arn in authorizer uri {uri}")
                continue

            api_gateway_lambdas_arns.add(lambda_arn)
        return api_gateway_lambdas_arns



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import SNS_TOPIC_TYPE


class SNSTopicGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = SNS_TOPIC_TYPE
    CONFIGURATION = {
        'region': None,
        "event_sources": list
    }



"""
    Copyright 2024 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
from abc import ABC, abstractmethod

from syndicate.commons import deep_get
from syndicate.commons.log_helper import get_user_logger
from syndicate.core.constants import EXPORT_DIR_NAME
from syndicate.core.helper import build_path

USER_LOG = get_user_logger()


class ConfigurationExporter(ABC):

    def __init__(self):
        from syndicate.core import CONFIG, RESOURCES_PROVIDER
        self.config = CONFIG
        self.project_path = self.config.project_path
        self.resources_provider = RESOURCES_PROVIDER

    def prepare_output_directory(self, output_dir):
        if not output_dir:
            output_dir_path = build_path(
                self.project_path, EXPORT_DIR_NAME)
        elif not os.path.isabs(output_dir):
            output_dir_path = build_path(
                os.getcwd(), output_dir)
        else:
            output_dir_path = output_dir
        os.makedirs(output_dir_path, exist_ok=True)
        return output_dir_path

    @staticmethod
    def _remove_prefix_suffix_from_string(string: str,
                                          prefix: str,
                                          suffix: str) -> str:
        if string.startswith(prefix) and prefix:
            string = string[len(prefix):]
        if string.endswith(suffix) and suffix:
            string = string[:-len(suffix)]
        return string

    @abstractmethod
    def export_configuration(self, *args, **kwargs):
        pass


class OASV3Exporter(ConfigurationExporter):

    def __init__(self):
        super().__init__()
        self.extended_prefix_mode = self.config.extended_prefix_mode
        self.prefix = self.config.resources_prefix
        self.suffix = self.config.resources_suffix

    def export_configuration(self, arn, meta):
        return self._export_api_gw_openapi_spec(arn, meta)

    def _export_api_gw_openapi_spec(self, api_arn: str,
                                    meta: dict) -> (str, dict):
        api_id = api_arn.split('/')[-1]
        api_stage = deep_get(meta, ['resource_meta', 'deploy_stage'])
        specification = self.resources_provider.api_gw().describe_openapi(
            api_id=api_id,
            stage_name=api_stage)
        if not specification:
            USER_LOG.warn(f'API Gateway not found by the ARN: "{api_arn}"')
            return api_id, None
        if self.extended_prefix_mode:
            specification['info']['title'] = (
                self._remove_prefix_suffix_from_string(
                    specification['info']['title'],
                    self.prefix,
                    self.suffix))
            self._remove_prefix_suffix_from_arn(
                specification,
                self.prefix,
                self.suffix)
        return api_id, specification

    def _remove_prefix_suffix_from_arn(self, spec: dict, prefix: str,
                                       suffix: str) -> None:
        def __remove_prefix_suffix(string: str, delimiter: str) -> str:
            if delimiter in string:
                origin_parts = string.split(delimiter)
                new_parts = []
                for part in origin_parts:
                    new_parts.append(
                        self._remove_prefix_suffix_from_string(
                            part, prefix, suffix))
                string = delimiter.join(new_parts)
            return string

        if isinstance(spec, dict):
            for key in spec.keys():
                if isinstance(spec[key], dict):
                    self._remove_prefix_suffix_from_arn(spec[key],
                                                        prefix, suffix)
                else:
                    if isinstance(spec[key], str):
                        if 'arn' in spec[key]:
                            spec[key] = __remove_prefix_suffix(spec[key], ':')
                            spec[key] = __remove_prefix_suffix(spec[key], '/')



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from functools import lru_cache

from syndicate.connection.api_gateway_connection import ApiGatewayConnection, \
    ApiGatewayV2Connection
from syndicate.connection.application_autoscaling_connection import (
    ApplicationAutoscaling)
from syndicate.connection.batch_connection import BatchConnection
from syndicate.connection.cloud_watch_connection import (EventConnection,
                                                         LogsConnection,
                                                         MetricConnection)
from syndicate.connection.cloudfront_connection import CloudFrontConnection
from syndicate.connection.cognito_identity_connection import (
    CognitoIdentityConnection)
from syndicate.connection.cognito_identity_provider_connection import \
    CognitoIdentityProviderConnection
from syndicate.connection.documentdb_connection import DocumentDBConnection
from syndicate.connection.dynamo_connection import DynamoConnection
from syndicate.connection.ec2_connection import EC2Connection
from syndicate.connection.elastic_beanstalk_connection import (
    BeanstalkConnection)
from syndicate.connection.firehose_connection import FirehoseConnection
from syndicate.connection.eventbridge_scheduler_connection import EventBridgeSchedulerConnection
from syndicate.connection.iam_connection import IAMConnection
from syndicate.connection.kinesis_connection import KinesisConnection
from syndicate.connection.kms_connection import KMSConnection
from syndicate.connection.lambda_connection import LambdaConnection
from syndicate.connection.s3_connection import S3Connection
from syndicate.connection.sns_connection import SNSConnection
from syndicate.connection.sqs_connection import SqsConnection
from syndicate.connection.step_functions_connection import SFConnection
from syndicate.connection.resource_groups_tagging_api_connection import \
    ResourceGroupsTaggingAPIConnection
from syndicate.connection.dax_connection import DaxConnection


class ConnectionProvider(object):
    def __init__(self, credentials):
        self.credentials = credentials.copy()

    @lru_cache(maxsize=None)
    def api_gateway(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return ApiGatewayConnection(**credentials)

    def api_gateway_v2(self, region=None):
        creds = self.credentials
        if region:
            creds = {**creds, 'region': region}
        return ApiGatewayV2Connection(**creds)

    @lru_cache(maxsize=None)
    def lambda_conn(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return LambdaConnection(**credentials)

    @lru_cache(maxsize=None)
    def cw_events(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return EventConnection(**credentials)

    @lru_cache(maxsize=None)
    def dynamodb(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return DynamoConnection(**credentials)

    @lru_cache(maxsize=None)
    def cognito_identity(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return CognitoIdentityConnection(**credentials)

    @lru_cache(maxsize=None)
    def cognito_identity_provider(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return CognitoIdentityProviderConnection(**credentials)

    @lru_cache(maxsize=None)
    def iam(self):
        return IAMConnection(**self.credentials)

    @lru_cache(maxsize=None)
    def s3(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return S3Connection(**credentials)

    @lru_cache(maxsize=None)
    def sns(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return SNSConnection(**credentials)

    @lru_cache(maxsize=None)
    def cw_logs(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return LogsConnection(**credentials)

    @lru_cache(maxsize=None)
    def cw_metric(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return MetricConnection(**credentials)

    @lru_cache(maxsize=None)
    def ec2(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return EC2Connection(**credentials)

    @lru_cache(maxsize=None)
    def cloud_front(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return CloudFrontConnection(**credentials)

    @lru_cache(maxsize=None)
    def beanstalk(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return BeanstalkConnection(**credentials)

    @lru_cache(maxsize=None)
    def step_functions(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return SFConnection(**credentials)

    @lru_cache(maxsize=None)
    def kinesis(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return KinesisConnection(**credentials)

    @lru_cache(maxsize=None)
    def firehose(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return FirehoseConnection(**credentials)

    @lru_cache(maxsize=None)
    def eventbridge_scheduler(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return EventBridgeSchedulerConnection(**credentials)

    @lru_cache(maxsize=None)
    def application_autoscaling(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return ApplicationAutoscaling(**credentials)

    @lru_cache(maxsize=None)
    def sqs(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return SqsConnection(**credentials)

    @lru_cache(maxsize=None)
    def kms(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return KMSConnection(**credentials)

    @lru_cache(maxsize=None)
    def batch(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return BatchConnection(**credentials)

    @lru_cache(maxsize=None)
    def documentdb(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return DocumentDBConnection(**credentials)

    @lru_cache(maxsize=None)
    def groups_tagging_api(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return ResourceGroupsTaggingAPIConnection(**credentials)

    @lru_cache(maxsize=None)
    def dax(self, region=None):
        credentials = self.credentials.copy()
        if region:
            credentials['region'] = region
        return DaxConnection(**credentials)



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from troposphere import kinesis

from syndicate.connection.kinesis_connection import validate_shard_count
from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import kinesis_stream_logic_name


class CfKinesisStreamConverter(CfResourceConverter):

    def convert(self, name, meta):
        stream = kinesis.Stream(kinesis_stream_logic_name(name))
        stream.Name = name
        shard_count = meta['shard_count']
        validate_shard_count(shard_count)
        stream.ShardCount = shard_count
        self.template.add_resource(stream)



from boto3 import client
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('syndicate.connection.firehose_connection')


@apply_methods_decorator(retry())
class FirehoseConnection(object):
    """ Firehose connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('firehose', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new Firehose connection.')

    def create_delivery_stream(self, stream_name, stream_type,
                               s3_configuration, kinesis_stream_source=None):
        params = {'DeliveryStreamName': stream_name,
                  'S3DestinationConfiguration': s3_configuration,
                  'DeliveryStreamType': stream_type}
        if kinesis_stream_source:
            params['KinesisStreamSourceConfiguration'] = kinesis_stream_source
        return self.client.create_delivery_stream(**params)['DeliveryStreamARN']

    def describe_delivery_stream(self, stream_name, limit=None,
                                 exclusive_start_id=None):
        params = {'DeliveryStreamName': stream_name}
        if limit:
            params['Limit'] = limit
        if exclusive_start_id:
            params['ExclusiveStartDestinationId'] = exclusive_start_id

        try:
            return self.client.describe_delivery_stream(**params)[
                    'DeliveryStreamDescription']
        except ClientError as e:
            if 'ResourceNotFoundException' in str(e):
                _LOG.warning(
                    f'Cannot find delivery stream with name {stream_name}')
                pass
            else:
                raise e

    def delete_delivery_stream(self, stream_name):
        try:
            return self.client.delete_delivery_stream(
                DeliveryStreamName=stream_name)
        except ClientError as e:
            if 'ResourceNotFoundException' in str(e):
                _LOG.warning(
                    f'Cannot find delivery stream with name {stream_name}')
                pass
            else:
                raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from boto3 import client
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('syndicate.connection.kinesis_connection')


def validate_shard_count(shard_count):
    if not isinstance(shard_count, int) or shard_count > 25:
        raise TypeError(
            'Shard count must be a valid integer '
            'less than 25 (max value per region). Actual type: {0}'.format(
                type(shard_count)))


@apply_methods_decorator(retry())
class KinesisConnection(object):
    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('kinesis', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new Kinesis connection.')

    def create_stream(self, stream_name, shard_count):
        validate_shard_count(shard_count)
        params = dict(StreamName=stream_name, ShardCount=shard_count)
        return self.client.create_stream(**params)

    def get_stream(self, stream_name):
        try:
            return self.client.describe_stream(StreamName=stream_name)[
                'StreamDescription']
        except ClientError as e:
            if 'ResourceNotFoundException' in str(e):
                pass  # valid exception
            else:
                raise e

    def get_list_streams(self):
        result = []
        response = self.client.list_streams()
        do_continue = response.get('HasMoreStreams')
        if 'StreamNames' in response:
            result.extend(response['StreamNames'])
        while do_continue:
            response = self.client.list_streams(
                ExclusiveStartStreamName=result[-1])
            do_continue = response.get('HasMoreStreams')
            if 'StreamNames' in response:
                result.extend(response['StreamNames'])
        return result

    def remove_stream(self, stream_name):
        self.client.delete_stream(StreamName=stream_name)
